"""The core functionality of CCL, including the core data types lives in this
module. Its focus is the :class:`Cosmology` class, which plays a central role,
carrying the information on cosmological parameters and derived quantities
needed in most of the calculations carried out by CCL.

.. note::
    All of the standalone functions in other modules, which take `cosmo` as
    their first argument, are methods of :class:`~Cosmology`.

Some important CCL parameters, governing for example the precision and speed
of some calculations, are independent of the :class:`~Cosmology` objects,
and instead can be accessed at a global level. You can do so as e.g.
``pyccl.gsl_params['ODE_GROWTH_EPSREL']``, ``pyccl.spline_params['K_MIN']``,
``pyccl.physical_constants['CLIGHT']``, or
``pyccl.gsl_params.ODE_GROWTH_EPSREL``, ``pyccl.spline_params.K_MIN``,
``pyccl.physical_constants.CLIGHT``.
"""
__all__ = ("TransferFunctions", "MatterPowerSpectra",
           "Cosmology", "CosmologyVanillaLCDM", "CosmologyCalculator",)

import yaml
from enum import Enum
from inspect import getmembers, isfunction, signature
from numbers import Real
from typing import Iterable

import numpy as np

from . import (
    CCLError, CCLObject, CCLParameters, CosmologyParams,
    DEFAULT_POWER_SPECTRUM, DefaultParams, Pk2D, check, lib,
    unlock_instance, emulators, baryons, modified_gravity)
from . import physical_constants as const


class TransferFunctions(Enum):
    BBKS = "bbks"
    EISENSTEIN_HU = "eisenstein_hu"
    EISENSTEIN_HU_NOWIGGLES = "eisenstein_hu_nowiggles"
    BOLTZMANN_CLASS = "boltzmann_class"
    BOLTZMANN_CAMB = "boltzmann_camb"
    BOLTZMANN_ISITGR = "boltzmann_isitgr"
    CALCULATOR = "calculator"
    EMULATOR_LINPK = "emulator"


class MatterPowerSpectra(Enum):
    LINEAR = "linear"
    HALOFIT = "halofit"
    CAMB = "camb"
    CALCULATOR = "calculator"
    EMULATOR_NLPK = "emulator"


# Configuration types
transfer_function_types = {
    'eisenstein_hu': lib.eisenstein_hu,
    'eisenstein_hu_nowiggles': lib.eisenstein_hu_nowiggles,
    'bbks': lib.bbks,
    'boltzmann_class': lib.boltzmann_class,
    'boltzmann_camb': lib.boltzmann_camb,
    'boltzmann_isitgr': lib.boltzmann_isitgr,
    'calculator': lib.pklin_from_input,
    'emulator': lib.emulator_linpk
}


matter_power_spectrum_types = {
    'halo_model': lib.halo_model,
    'halofit': lib.halofit,
    'linear': lib.linear,
    'calculator': lib.pknl_from_input,
    'camb': lib.pknl_from_boltzman,
    'emulator': lib.emulator_nlpk
}

_TOP_LEVEL_MODULES = ("",)


def _make_methods(cls=None, *, modules=_TOP_LEVEL_MODULES, name=None):
    """Assign all functions in ``modules`` which take ``name`` as their
    first argument as methods of the class ``cls``.
    """
    import functools
    from importlib import import_module

    if cls is None:
        # called with parentheses
        return functools.partial(_make_methods, modules=modules)

    pkg = __name__.rsplit(".")[0]
    modules = [import_module(f".{module}", pkg) for module in modules]
    funcs = [getmembers(module, isfunction) for module in modules]
    funcs = [func for sublist in funcs for func in sublist]

    for name, func in funcs:
        pars = signature(func).parameters
        if pars and list(pars)[0] == "cosmo":
            setattr(cls, name, func)

    return cls


@_make_methods(modules=("", "halos", "nl_pt",), name="cosmo")
class Cosmology(CCLObject):
    """Stores information about cosmological parameters and associated data
    (e.g. distances, power spectra).

    The values of cosmological parameters may be looked up by name
    (e.g. ``cosmo["sigma8"]``). Note that some of the parameters accessible
    this way are not contained in the signature of :class:`~Cosmology`, but
    are derived during initialization.

    .. note:: Although some arguments default to `None`, they will raise a
              ValueError inside this function if not specified, so they are not
              optional.

    .. note:: The parameter ``Omega_g`` can be used to set the radiation density
              (not including relativistic neutrinos) to zero. Doing this will
              give you a model that is physically inconsistent since the
              temperature of the CMB will still be non-zero.

    Args:
        Omega_c (:obj:`float`): Cold dark matter density fraction.
        Omega_b (:obj:`float`): Baryonic matter density fraction.
        h (:obj:`float`): Hubble constant divided by 100 km/s/Mpc; unitless.
        A_s (:obj:`float`): Power spectrum normalization. Exactly one of A_s
            and sigma_8 is required.
        sigma8 (:obj:`float`): Variance of matter density perturbations at
            an 8 Mpc/h scale. Exactly one of A_s and sigma_8 is required.
            Note that, if a value of `sigma8` is passed, CCL will enforce
            the linear matter power spectrum to be correctly normalised to
            this value of :math:`\\sigma_8`, even in the presence of other
            parameters (e.g. modified gravity parameters) that might affect
            the overall power spectrum normalization.
        n_s (:obj:`float`): Primordial scalar perturbation spectral index.
        Omega_k (:obj:`float`): Curvature density fraction.
            Defaults to 0.
        Omega_g (:obj:`float`): Density in relativistic species
            except massless neutrinos. The default of `None` corresponds
            to setting this from the CMB temperature. Note that if a non-`None`
            value is given, this may result in a physically inconsistent model
            because the CMB temperature will still be non-zero in the
            parameters.
        Neff (:obj:`float`): Effective number of massless
            neutrinos present. Defaults to 3.044.
        m_nu (:obj:`float` or `array`):
            Mass in eV of the massive neutrinos present. Defaults to 0.
            If a sequence is passed, it is assumed that the elements of the
            sequence represent the individual neutrino masses.
        mass_split (:obj:`str`): Type of massive neutrinos. Should
            be one of 'single', 'equal', 'normal', 'inverted'. 'single' treats
            the mass as being held by one massive neutrino. The other options
            split the mass into 3 massive neutrinos. Ignored if a sequence is
            passed in ``m_nu``. Default is 'normal'.
        w0 (:obj:`float`): First order term of dark energy equation
            of state. Defaults to -1.
        wa (:obj:`float`): Second order term of dark energy equation
            of state. Defaults to 0.
        T_CMB (:obj:`float`): The CMB temperature today. The default value
            is 2.7255.
        T_ncdm (:obj:`float`): Non-CDM temperature in units of photon
            temperature. The default is 0.71611.
        transfer_function (:obj:`str` or :class:`~pyccl.emulators.emu_base.EmulatorPk`):
            The transfer function to use. Defaults to 'boltzmann_camb'.
        matter_power_spectrum (:obj:`str` or :class:`~pyccl.emulators.emu_base.EmulatorPk`):
            The matter power spectrum to use. Defaults to 'halofit'.
        baryonic_effects (:class:`~pyccl.baryons.baryons_base.Baryons` or :obj:`None`):
            The baryonic effects model to use. Options are `None` (no baryonic effects), or
            a :class:`~pyccl.baryons.baryons_base.Baryons` object.
        mg_parametrization (:class:`~pyccl.modified_gravity.modified_gravity_base.ModifiedGravity` or `None`):
            The modified gravity parametrization to use. Options are `None`
            (no MG), or a :class:`~pyccl.modified_gravity.modified_gravity_base.ModifiedGravity`
            object. Currently, only :class:`~pyccl.modified_gravity.mu_Sigma.MuSigmaMG`
            is supported.
        extra_parameters (:obj:`dict`): Dictionary holding extra
            parameters. Currently supports extra parameters for CAMB.
            Details described below. Defaults to None.

    Currently supported extra parameters for CAMB are:

        * `halofit_version`
        * `HMCode_A_baryon`
        * `HMCode_eta_baryon`
        * `HMCode_logT_AGN`
        * `kmax`
        * `lmax`
        * `dark_energy_model`

    Consult the CAMB documentation for their usage. These parameters are passed
    in a :obj:`dict` to `extra_parameters` as::

        extra_parameters = {"camb": {"halofit_version": "mead2020_feedback",
                                     "HMCode_logT_AGN": 7.8}}

    .. note :: If using camb to compute the non-linear power spectrum with HMCode
               to include baryonic effects, you should not include any extra
               baryonic effects (i.e. set `baryonic_effects=None`).
    """ # noqa
    from ._core.repr_ import build_string_Cosmology as __repr__
    __eq_attrs__ = ("_params_init_kwargs", "_config_init_kwargs",
                    "_accuracy_params", "lin_pk_emu", 'nl_pk_emu',
                    "baryons", "mg_parametrization")

    def __init__(
            self, *, Omega_c=None, Omega_b=None, h=None, n_s=None,
            sigma8=None, A_s=None, Omega_k=0., Omega_g=None,
            Neff=None, m_nu=0., mass_split='normal', w0=-1., wa=0.,
            T_CMB=DefaultParams.T_CMB,
            T_ncdm=DefaultParams.T_ncdm,
            transfer_function='boltzmann_camb',
            matter_power_spectrum='halofit',
            baryonic_effects=None,
            mg_parametrization=None,
            extra_parameters=None):

        if Neff is None:
            Neff = 3.044

        extra_parameters = extra_parameters or {}

        # initialise linear Pk emulators if needed
        self.lin_pk_emu = None
        if isinstance(transfer_function, emulators.EmulatorPk):
            self.lin_pk_emu = transfer_function
            transfer_function = 'emulator'

        # initialise nonlinear Pk emulators if needed
        self.nl_pk_emu = None
        if isinstance(matter_power_spectrum, emulators.EmulatorPk):
            self.nl_pk_emu = matter_power_spectrum
            matter_power_spectrum = 'emulator'

        self.baryons = baryonic_effects
        if not isinstance(self.baryons, baryons.Baryons):
            if self.baryons is not None:
                raise ValueError("`baryonic_effects` must be `None` "
                                 "or a `Baryons` instance.")

        self.mg_parametrization = mg_parametrization
        if self.mg_parametrization is not None and not isinstance(
                self.mg_parametrization,
                modified_gravity.ModifiedGravity):
            raise ValueError("`mg_parametrization` must be `None` "
                             "or a `ModifiedGravity` instance.")

        if self.mg_parametrization is None:
            # Internally, CCL still relies exclusively on the mu-Sigma
            # parametrization, so we fill that in for now unless something
            # else is provided.
            self.mg_parametrization = modified_gravity.MuSigmaMG()
        if not isinstance(
                self.mg_parametrization,
                modified_gravity.MuSigmaMG):
            raise NotImplementedError("`mg_parametrization` only supports the "
                                      "mu-Sigma parametrization at this point")

        # going to save these for later
        self._params_init_kwargs = dict(
            Omega_c=Omega_c, Omega_b=Omega_b, h=h, n_s=n_s, sigma8=sigma8,
            A_s=A_s, Omega_k=Omega_k, Omega_g=Omega_g, Neff=Neff, m_nu=m_nu,
            mass_split=mass_split, w0=w0, wa=wa, T_CMB=T_CMB, T_ncdm=T_ncdm,
            extra_parameters=extra_parameters)

        self._config_init_kwargs = dict(
            transfer_function=transfer_function,
            matter_power_spectrum=matter_power_spectrum,
            extra_parameters=extra_parameters)

        self._build_cosmo()

        self._pk_lin = {}
        self._pk_nl = {}

    def _build_cosmo(self):
        """Assemble all of the input data into a valid ccl_cosmology object."""
        # We have to make all of the C stuff that goes into a cosmology
        # and then we make the cosmology.
        self._build_parameters(**self._params_init_kwargs)
        self._build_config(**self._config_init_kwargs)
        self.cosmo = lib.cosmology_create(self._params, self._config)
        self._spline_params = CCLParameters.get_params_dict("spline_params")
        self._gsl_params = CCLParameters.get_params_dict("gsl_params")
        self._accuracy_params = {**self._spline_params, **self._gsl_params}

        if self.cosmo.status != 0:
            raise CCLError(f"{self.cosmo.status}: {self.cosmo.status_message}")

    def write_yaml(self, filename, *, sort_keys=False):
        """Write a YAML representation of the parameters to file.

        Args:
            filename (:obj:`str`): file name, file pointer, or stream to write
                parameters to.
        """
        def make_yaml_friendly(d):
            # serialize numpy types and dicts
            for k, v in d.items():
                if isinstance(v, int):
                    d[k] = int(v)
                elif isinstance(v, float):
                    d[k] = float(v)
                elif isinstance(v, dict):
                    make_yaml_friendly(v)

        params = {**self._params_init_kwargs, **self._config_init_kwargs}
        make_yaml_friendly(params)

        if isinstance(filename, str):
            with open(filename, "w") as fp:
                return yaml.dump(params, fp, sort_keys=sort_keys)
        return yaml.dump(params, filename, sort_keys=sort_keys)

    @classmethod
    def read_yaml(cls, filename, **kwargs):
        """Read the parameters from a YAML file.

        Args:
            filename (:obj:`str`): file name, file pointer, or stream to read
                parameters from.
            **kwargs (:obj:`dict`): additional keywords that supersede
                file contents
        """
        loader = yaml.Loader
        if isinstance(filename, str):
            with open(filename, 'r') as fp:
                return cls(**{**yaml.load(fp, Loader=loader), **kwargs})
        return cls(**{**yaml.load(filename, Loader=loader), **kwargs})

    def _build_config(
            self, transfer_function=None, matter_power_spectrum=None,
            extra_parameters=None):
        """Build a ccl_configuration struct.

        This function builds C ccl_configuration struct. This structure
        controls which various approximations are used for the transfer
        function, matter power spectrum, and baryonic effect in the matter
        power spectrum.

        It also does some error checking on the inputs to make sure they
        are valid and physically consistent.
        """
        if (matter_power_spectrum == "camb"
                and transfer_function != "boltzmann_camb"):
            raise CCLError(
                "To compute the non-linear matter power spectrum with CAMB "
                "the transfer function should be 'boltzmann_camb'.")

        config = lib.configuration()
        tf = transfer_function_types[transfer_function]
        config.transfer_function_method = tf
        mps = matter_power_spectrum_types[matter_power_spectrum]
        config.matter_power_spectrum_method = mps

        # Store ccl_configuration for later access
        self._config = config

    def _build_parameters(
            self, Omega_c=None, Omega_b=None, h=None, n_s=None, sigma8=None,
            A_s=None, Omega_k=None, Neff=None, m_nu=None, mass_split=None,
            w0=None, wa=None, T_CMB=None, T_ncdm=None,
            Omega_g=None, extra_parameters=None):
        """Build a ccl_parameters struct"""
        # Fill-in defaults (SWIG converts `numpy.nan` to `NAN`)
        A_s = np.nan if A_s is None else A_s
        sigma8 = np.nan if sigma8 is None else sigma8
        Omega_g = np.nan if Omega_g is None else Omega_g

        # Check to make sure Omega_k is within reasonable bounds.
        if Omega_k < -1.0135:
            raise ValueError("Omega_k must be more than -1.0135.")

        # Check to make sure specified amplitude parameter is consistent.
        if [A_s, sigma8].count(np.nan) != 1:
            raise ValueError("Set either A_s or sigma8 but not both.")

        # Check if any compulsory parameters are not set.
        compul = {"Omega_c": Omega_c, "Omega_b": Omega_b, "h": h, "n_s": n_s}
        for param, value in compul.items():
            if value is None:
                raise ValueError(f"Must set parameter {param}.")

        # Make sure the neutrino parameters are consistent.
        if not isinstance(m_nu, (Real, Iterable)):
            raise ValueError("m_nu must be float or sequence")

        c = const
        # Initialize curvature.
        k_sign = -np.sign(Omega_k) if np.abs(Omega_k) > 1e-6 else 0
        sqrtk = np.sqrt(np.abs(Omega_k)) * h / c.CLIGHT_HMPC

        # Initialize radiation.
        rho_g = 4 * c.STBOLTZ / c.CLIGHT**3 * T_CMB**4
        rho_crit = c.RHO_CRITICAL * c.SOLAR_MASS / c.MPC_TO_METER**3 * h**2

        # Initialize neutrinos.
        g = (4/11)**(1/3)
        T_nu = g * T_CMB
        massless_limit = T_nu * c.KBOLTZ / c.EV_IN_J

        from .neutrinos import nu_masses
        mnu_list = nu_masses(m_nu=m_nu, mass_split=mass_split)
        nu_mass = mnu_list[mnu_list > massless_limit]
        N_nu_mass = len(nu_mass)
        N_nu_rel = Neff - N_nu_mass * (T_ncdm/g)**4

        rho_nu_rel = N_nu_rel * (7/8) * 4 * c.STBOLTZ / c.CLIGHT**3 * T_nu**4
        Omega_nu_rel = rho_nu_rel / rho_crit
        Omega_nu_mass = self._OmNuh2(nu_mass, N_nu_mass, T_CMB, T_ncdm) / h**2

        if N_nu_rel < 0:
            raise ValueError("Unphysical Neff and m_nu combination results to "
                             "negative number of relativistic neutrinos.")

        # Initialize matter & dark energy.
        Omega_m = Omega_b + Omega_c + Omega_nu_mass
        Omega_l = 1 - Omega_m - rho_g/rho_crit - Omega_nu_rel - Omega_k
        if np.isnan(Omega_g):
            # No value passed for Omega_g
            Omega_g = rho_g/rho_crit
        else:
            # Omega_g was passed - modify Omega_l
            Omega_l += rho_g/rho_crit - Omega_g

        # Take the mu-Sigma parameters from the modified_gravity container
        # object. This is the only supported MG parametrization at this time.
        assert isinstance(self.mg_parametrization, modified_gravity.MuSigmaMG)
        mu_0 = self.mg_parametrization.mu_0
        sigma_0 = self.mg_parametrization.sigma_0
        c1_mg = self.mg_parametrization.c1_mg
        c2_mg = self.mg_parametrization.c2_mg
        lambda_mg = self.mg_parametrization.lambda_mg

        self._fill_params(
            m_nu=nu_mass, sum_nu_masses=sum(nu_mass), N_nu_mass=N_nu_mass,
            N_nu_rel=N_nu_rel, Neff=Neff, Omega_nu_mass=Omega_nu_mass,
            Omega_nu_rel=Omega_nu_rel, Omega_m=Omega_m, Omega_c=Omega_c,
            Omega_b=Omega_b, Omega_k=Omega_k, sqrtk=sqrtk, k_sign=int(k_sign),
            T_CMB=T_CMB, T_ncdm=T_ncdm, Omega_g=Omega_g, w0=w0, wa=wa,
            Omega_l=Omega_l, h=h, H0=h*100, A_s=A_s, sigma8=sigma8, n_s=n_s,
            mu_0=mu_0, sigma_0=sigma_0, c1_mg=c1_mg, c2_mg=c2_mg,
            lambda_mg=lambda_mg)

    def _OmNuh2(self, m_nu, N_nu_mass, T_CMB, T_ncdm):
        # Compute OmNuh2 today.
        ret, st = lib.Omeganuh2_vec(N_nu_mass, T_CMB, T_ncdm, [1], m_nu, 1, 0)
        check(st)
        return ret[0]

    def _fill_params(self, **kwargs):
        if not hasattr(self, "_params"):
            self._params = CosmologyParams()
        [setattr(self._params, par, val) for par, val in kwargs.items()]

    def __getitem__(self, key):
        if key == 'extra_parameters':
            return self._params_init_kwargs["extra_parameters"]
        return getattr(self._params, key)

    def __del__(self):
        """Free the C memory this object is managing as it is being garbage
        collected (hopefully)."""
        if hasattr(self, "cosmo"):
            lib.cosmology_free(self.cosmo)
            delattr(self, "cosmo")
        if hasattr(self, "_params"):
            lib.parameters_free(self._params)
            delattr(self, "_params")

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        """Free the C memory this object is managing when the context manager
        exits."""
        self.__del__()

    def __getstate__(self):
        # we are removing any C data before pickling so that the
        # is pure python when pickled.
        state = self.__dict__.copy()
        state.pop('cosmo', None)
        state.pop('_params', None)
        state.pop('_config', None)
        return state

    def __setstate__(self, state):
        # This will create a new `Cosmology` object so we create another lock.
        state["_object_lock"] = type(state.pop("_object_lock"))()
        self.__dict__ = state
        # we removed the C data when it was pickled, so now we unpickle
        # and rebuild the C data
        self._build_cosmo()
        self._object_lock.lock()  # Lock on exit.

    def compute_distances(self):
        """Compute the distance splines."""
        if self.has_distances:
            return
        status = 0
        status = lib.cosmology_compute_distances(self.cosmo, status)
        check(status, self)

    def compute_growth(self):
        """Compute the growth function."""
        if self.has_growth:
            return
        status = 0
        status = lib.cosmology_compute_growth(self.cosmo, status)
        check(status, self)

    def _compute_linear_power(self):
        """Return the linear power spectrum."""
        self.compute_growth()

        # Populate power spectrum splines
        trf = self._config_init_kwargs['transfer_function']
        pk = None
        rescale_s8 = True
        rescale_mg = True
        if trf == "boltzmann_camb":
            rescale_s8 = False
            # For MG, the input sigma8 includes the effects of MG, while the
            # sigma8 that CAMB uses is the GR definition. So we need to rescale
            # sigma8 afterwards.
            if self.mg_parametrization.mu_0 != 0:
                rescale_s8 = True
        elif trf == 'boltzmann_class':
            pk = self.get_class_pk_lin()
        elif trf == 'boltzmann_isitgr':
            rescale_mg = False
            pk = self.get_isitgr_pk_lin()
        elif trf in ['bbks', 'eisenstein_hu', 'eisenstein_hu_nowiggles']:
            rescale_s8 = False
            rescale_mg = False
            pk = Pk2D.from_model(self, model=trf)
        elif trf == 'emulator':
            rescale_s8 = False
            pk = self.lin_pk_emu.get_pk2d(self)

        # Compute the CAMB nonlin power spectrum if needed,
        # to avoid repeating the code in `compute_nonlin_power`.
        # Because CAMB power spectra come in pairs with pkl always computed,
        # we set the nonlin power spectrum first, but keep the linear via a
        # status variable to use it later if the transfer function is CAMB too.
        pkl = None
        if self._config_init_kwargs["matter_power_spectrum"] == "camb":
            rescale_mg = False
            if self.mg_parametrization.mu_0 != 0:
                raise ValueError("Can't rescale non-linear power spectrum "
                                 "from CAMB for mu-Sigma MG.")
            name = "delta_matter:delta_matter"
            pkl, self._pk_nl[name] = self.get_camb_pk_lin(nonlin=True)

        if trf == "boltzmann_camb":
            pk = pkl if pkl is not None else self.get_camb_pk_lin()

        # Rescale by sigma8/mu-sigma if needed
        if pk:
            status = 0
            status = lib.rescale_linpower(self.cosmo, pk.psp,
                                          int(rescale_mg),
                                          int(rescale_s8),
                                          status)
            check(status, self)

        return pk

    @unlock_instance(mutate=False)
    def compute_linear_power(self):
        """Compute the linear power spectrum."""
        if self.has_linear_power:
            return
        self._pk_lin[DEFAULT_POWER_SPECTRUM] = self._compute_linear_power()

    def _compute_nonlin_power(self):
        """Return the non-linear power spectrum."""
        self.compute_distances()

        # Populate power spectrum splines
        mps = self._config_init_kwargs['matter_power_spectrum']
        # needed for halofit, and linear options
        if (mps not in ['emulator']) and (mps is not None):
            self.compute_linear_power()

        if mps == "camb" and self.has_nonlin_power:
            # Already computed
            return self._pk_nl[DEFAULT_POWER_SPECTRUM]

        if mps == 'halofit':
            pkl = self._pk_lin[DEFAULT_POWER_SPECTRUM]
            pk = pkl.apply_halofit(self)
        elif mps == 'linear':
            pk = self._pk_lin[DEFAULT_POWER_SPECTRUM]
        elif mps == 'emulator':
            pk = self.nl_pk_emu.get_pk2d(self)

        # Include baryonic effects
        if self.baryons is not None:
            pk = self.baryons.include_baryonic_effects(self, pk)
        return pk

    @unlock_instance(mutate=False)
    def compute_nonlin_power(self):
        """Compute the non-linear power spectrum."""
        if self.has_nonlin_power:
            return
        self._pk_nl[DEFAULT_POWER_SPECTRUM] = self._compute_nonlin_power()

    def compute_sigma(self):
        """Compute the sigma(M) spline."""
        if self.has_sigma:
            return

        pk = self.get_linear_power()
        status = 0
        status = lib.cosmology_compute_sigma(self.cosmo, pk.psp, status)
        check(status, self)

    def get_linear_power(self, name=DEFAULT_POWER_SPECTRUM):
        """Get the :class:`~pyccl.pk2d.Pk2D` object associated with
        the linear power spectrum with name ``name``.

        Args:
            name (:obj:`str` or :obj:`None`): name of the power spectrum to
                return.

        Returns:
            :class:`~pyccl.pk2d.Pk2D` object containing the linear
            power spectrum with name `name`.
        """
        if name == DEFAULT_POWER_SPECTRUM:
            self.compute_linear_power()
        pk = self._pk_lin.get(name)
        if pk is None:
            raise KeyError(f"Power spectrum {name} does not exist.")
        return pk

    def get_nonlin_power(self, name=DEFAULT_POWER_SPECTRUM):
        """Get the :class:`~pyccl.pk2d.Pk2D` object associated with
        the non-linear power spectrum with name ``name``.

        Args:
            name (:obj:`str` or :obj:`None`): name of the power spectrum to
                return.

        Returns:
            :class:`~pyccl.pk2d.Pk2D` object containing the non-linear
            power spectrum with name ``name``.
        """
        if name == DEFAULT_POWER_SPECTRUM:
            self.compute_nonlin_power()
        pk = self._pk_nl.get(name)
        if pk is None:
            raise KeyError(f"Power spectrum {name} does not exist.")
        return pk

    @property
    def has_distances(self):
        """Checks if the distances have been precomputed."""
        return bool(self.cosmo.computed_distances)

    @property
    def has_growth(self):
        """Checks if the growth function has been precomputed."""
        return bool(self.cosmo.computed_growth)

    @property
    def has_linear_power(self):
        """Checks if the linear power spectra have been precomputed."""
        return DEFAULT_POWER_SPECTRUM in self._pk_lin

    @property
    def has_nonlin_power(self):
        """Checks if the non-linear power spectra have been precomputed."""
        return DEFAULT_POWER_SPECTRUM in self._pk_nl

    @property
    def has_sigma(self):
        """Checks if sigma(M) is precomputed."""
        return bool(self.cosmo.computed_sigma)


def CosmologyVanillaLCDM(**kwargs):
    """A cosmology with typical flat Lambda-CDM parameters (`Omega_c=0.25`,
    `Omega_b = 0.05`, `Omega_k = 0`, `sigma8 = 0.81`, `n_s = 0.96`, `h = 0.67`,
    no massive neutrinos) for quick instantiation.

    Arguments:
        **kwargs (:obj:`dict`): a dictionary of parameters passed as arguments
            to the :class:`Cosmology` constructor. It should not contain any of
            the :math:`\\Lambda`-CDM parameters (`"Omega_c"`, `"Omega_b"`,
            `"n_s"`, `"sigma8"`, `"A_s"`, `"h"`), since these are fixed.
    """
    p = {'Omega_c': 0.25,
         'Omega_b': 0.05,
         'h': 0.67,
         'n_s': 0.96,
         'sigma8': 0.81,
         'A_s': None}
    if set(p).intersection(set(kwargs)):
        raise ValueError(
            f"You cannot change the ΛCDM parameters: {list(p.keys())}.")
    # TODO py39+: dictionary union operator `(p | kwargs)`.
    return Cosmology(**{**p, **kwargs})


class CosmologyCalculator(Cosmology):
    """A "calculator-mode" CCL :class:`~Cosmology` object.
    This allows users to build a cosmology from a set of arrays
    describing the background expansion, linear growth factor and
    linear and non-linear power spectra, which can then be used
    to compute more complex observables (e.g. angular power
    spectra or halo-model quantities). These are stored in
    ``background``, ``growth``, ``pk_linear`` and ``pk_nonlin``.

    .. note:: Although in principle these arrays should suffice
              to compute most observable quantities some
              calculations implemented in CCL (e.g. the halo
              mass function) requires knowledge of basic
              cosmological parameters such as :math:`\\Omega_M`.
              For this reason, users must pass a minimal set
              of :math:`\\Lambda`-CDM cosmological parameters.

    Args:
        Omega_c (:obj:`float`): Cold dark matter density fraction.
        Omega_b (:obj:`float`): Baryonic matter density fraction.
        h (:obj:`float`): Hubble constant divided by 100 km/s/Mpc; unitless.
        A_s (:obj:`float`): Power spectrum normalization. Exactly one of A_s
            and sigma_8 is required.
        sigma8 (:obj:`float`): Variance of matter density perturbations at
            an 8 Mpc/h scale. Exactly one of A_s and sigma_8 is required.
        n_s (:obj:`float`): Primordial scalar perturbation spectral index.
        Omega_k (:obj:`float`): Curvature density fraction.
            Defaults to 0.
        Omega_g (:obj:`float`): Density in relativistic species
            except massless neutrinos. The default of `None` corresponds
            to setting this from the CMB temperature. Note that if a non-`None`
            value is given, this may result in a physically inconsistent model
            because the CMB temperature will still be non-zero in the
            parameters.
        Neff (:obj:`float`): Effective number of massless
            neutrinos present. Defaults to 3.044.
        m_nu (:obj:`float` or `array`):
            Mass in eV of the massive neutrinos present. Defaults to 0.
            If a sequence is passed, it is assumed that the elements of the
            sequence represent the individual neutrino masses.
        mass_split (:obj:`str`): Type of massive neutrinos. Should
            be one of 'single', 'equal', 'normal', 'inverted'. 'single' treats
            the mass as being held by one massive neutrino. The other options
            split the mass into 3 massive neutrinos. Ignored if a sequence is
            passed in ``m_nu``. Default is 'normal'.
        w0 (:obj:`float`): First order term of dark energy equation
            of state. Defaults to -1.
        wa (:obj:`float`): Second order term of dark energy equation
            of state. Defaults to 0.
        T_CMB (:obj:`float`): The CMB temperature today. The default value
            is 2.7255.
        T_ncdm (:obj:`float`): Non-CDM temperature in units of photon
            temperature. The default is the same as in the base class
        mg_parametrization (:class:`~pyccl.modified_gravity.modified_gravity_base.ModifiedGravity` or `None`):
            The modified gravity parametrization to use. Options are `None`
            (no MG), or a :class:`~pyccl.modified_gravity.modified_gravity_base.ModifiedGravity`
            object. Currently, only :class:`~pyccl.modified_gravity.mu_Sigma.MuSigmaMG`
            is supported.
        background (:obj:`dict`): a dictionary describing the background
            expansion. It must contain three mandatory entries: ``'a'``: an
            array of monotonically ascending scale-factor values. ``'chi'``:
            an array containing the values of the comoving radial distance
            (in units of Mpc) at the scale factor values stored in ``a``.
            '``h_over_h0``': an array containing the Hubble expansion rate at
            the scale factor values stored in ``a``, divided by its value
            today (at ``a=1``).
        growth (:obj:`dict`): a dictionary describing the linear growth of
            matter fluctuations. It must contain three mandatory entries:
            ``'a'``: an array of monotonically ascending scale-factor
            values. ``'growth_factor'``: an array containing the values of
            the linear growth factor :math:`D(a)` at the scale factor
            values stored in ``a``. '``growth_rate``': an array containing the
            growth rate :math:`f(a)\\equiv d\\log D/d\\log a` at the scale
            factor values stored in ``a``.
        pk_linear (:obj:`dict`): a dictionary containing linear power
            spectra. It must contain the following mandatory entries:
            ``'a'``: an array of scale factor values. ``'k'``: an array of
            comoving wavenumbers in units of inverse Mpc.
            ``'delta_matter:delta_matter'``: a 2D array of shape
            ``(n_a, n_k)``, where ``n_a`` and ``n_k`` are the lengths of
            ``'a'`` and ``'k'`` respectively, containing the linear matter
            power spectrum :math:`P(k,a)`. This dictionary may also
            contain other entries with keys of the form ``'q1:q2'``,
            containing other cross-power spectra between quantities
            ``'q1'`` and ``'q2'``.
        pk_nonlin (:obj:`dict`): a dictionary containing non-linear
            power spectra. It must contain the following mandatory
            entries: ``'a'``: an array of scale factor values.
            ``'k'``: an array of comoving wavenumbers in units of
            inverse Mpc. If ``nonlinear_model`` is ``None``, it should also
            contain ``'delta_matter:delta_matter'``: a 2D array of
            shape ``(n_a, n_k)``, where ``n_a`` and ``n_k`` are the lengths
            of ``'a'`` and ``'k'`` respectively, containing the non-linear
            matter power spectrum :math:`P(k,a)`. This dictionary may
            also contain other entries with keys of the form ``'q1:q2'``,
            containing other cross-power spectra between quantities
            ``'q1'`` and ``'q2'``.
        nonlinear_model (:obj:`str`, :obj:`dict` or :obj:`None`): model to
            compute non-linear power spectra. If a string, the associated
            non-linear model will be applied to all entries in ``pk_linear``
            which do not appear in ``pk_nonlin``. If a dictionary, it should
            contain entries of the form ``'q1:q2': model``, where ``model``
            is a string designating the non-linear model to apply to the
            ``'q1:q2'`` power spectrum, which must also be present in
            ``pk_linear``. If ``model`` is ``None``, this non-linear power
            spectrum will not be calculated. If ``nonlinear_model`` is
            ``None``, no additional non-linear power spectra will be
            computed. The only non-linear model supported is ``'halofit'``,
            corresponding to the "HALOFIT" transformation of
            `Takahashi et al. 2012 <https://arxiv.org/abs/1208.2701>`_.
    """ # noqa
    __eq_attrs__ = ("_params_init_kwargs", "_config_init_kwargs",
                    "_accuracy_params", "_input_arrays",)

    def __init__(
            self, *, Omega_c=None, Omega_b=None, h=None, n_s=None,
            sigma8=None, A_s=None, Omega_k=0., Omega_g=None,
            Neff=None, m_nu=0., mass_split="normal", w0=-1., wa=0.,
            T_CMB=DefaultParams.T_CMB, T_ncdm=DefaultParams.T_ncdm,
            mg_parametrization=None, background=None, growth=None,
            pk_linear=None, pk_nonlin=None, nonlinear_model=None):

        super().__init__(
            Omega_c=Omega_c, Omega_b=Omega_b, h=h, n_s=n_s, sigma8=sigma8,
            A_s=A_s, Omega_k=Omega_k, Omega_g=Omega_g, Neff=Neff, m_nu=m_nu,
            mass_split=mass_split, w0=w0, wa=wa, T_CMB=T_CMB, T_ncdm=T_ncdm,
            mg_parametrization=mg_parametrization,
            transfer_function="calculator", matter_power_spectrum="calculator")

        self._input_arrays = {"background": background, "growth": growth,
                              "pk_linear": pk_linear, "pk_nonlin": pk_nonlin,
                              "nonlinear_model": nonlinear_model}

        if background is not None:
            self._init_background(background)
        if growth is not None:
            self._init_growth(growth)
        if pk_linear is not None:
            self._init_pk_linear(pk_linear)
        if pk_nonlin is not None:
            self._init_pk_nonlinear(pk_nonlin, nonlinear_model)
        self._apply_nonlinear_model(nonlinear_model)

    def _check_scale_factor(self, a):
        if not (np.diff(a) > 0).all():
            raise ValueError("Scale factor not monotonically increasing.")
        if np.abs(a[-1] - 1) > 1e-5:
            raise ValueError("Scale factor should end at 1.")

    def _check_input(self, a, arr1, arr2):
        self._check_scale_factor(a)
        if not a.shape == arr1.shape == arr2.shape:
            raise ValueError("Shape mismatch of input arrays.")

    def _check_label(self, name):
        if len(name.split(":")) != 2:
            raise ValueError(f"Could not parse power spectrum {name}. "
                             "Label must be of the form 'q1:q2'.")

    def _init_background(self, background):
        a, chi, E = background["a"], background["chi"], background["h_over_h0"]
        self._check_input(a, chi, E)
        status = 0
        status = lib.cosmology_distances_from_input(self.cosmo, a, chi, E,
                                                    status)
        check(status, self)

    def _init_growth(self, growth):
        a, gz, fz = growth["a"], growth["growth_factor"], growth["growth_rate"]
        self._check_input(a, gz, fz)
        status = 0
        status = lib.cosmology_growth_from_input(self.cosmo, a, gz, fz, status)
        check(status, self)

    def _init_pk_linear(self, pk_linear):
        a, lk = pk_linear["a"], np.log(pk_linear["k"])
        self._check_scale_factor(a)
        self.compute_growth()  # needed for high-z extrapolation
        na, nk = a.size, lk.size

        if DEFAULT_POWER_SPECTRUM not in pk_linear:
            raise ValueError("pk_linear does not contain "
                             f"{DEFAULT_POWER_SPECTRUM}")

        pk_names = set(pk_linear.keys()) - set(["a", "k"])
        for name in pk_names:
            self._check_label(name)
            pk = pk_linear[name]
            if pk.shape != (na, nk):
                raise ValueError("Power spectrum shape mismatch. "
                                 f"Expected {(na, nk)}. Received {pk.shape}.")
            # Spline in log-space if the P(k) is positive-definite
            use_log = (pk > 0).all()
            if use_log:
                pk = np.log(pk)
            self._pk_lin[name] = Pk2D(a_arr=a, lk_arr=lk, pk_arr=pk,
                                      is_logp=use_log)

    def _init_pk_nonlinear(self, pk_nonlin, nonlinear_model):
        a, lk = pk_nonlin["a"], np.log(pk_nonlin["k"])
        self._check_scale_factor(a)
        na, nk = a.size, lk.size

        if DEFAULT_POWER_SPECTRUM not in pk_nonlin and nonlinear_model is None:
            raise ValueError(f"{DEFAULT_POWER_SPECTRUM} not specified in "
                             "`pk_nonlin` and `nonlinear_model` is None")

        pk_names = set(pk_nonlin.keys()) - set(["a", "k"])
        for name in pk_names:
            self._check_label(name)
            pk = pk_nonlin[name]
            if pk.shape != (na, nk):
                raise ValueError("Power spectrum shape mismatch. "
                                 f"Expected {(na, nk)}. Received {pk.shape}.")
            # Spline in log-space if the P(k) is positive-definite
            use_log = (pk > 0).all()
            if use_log:
                pk = np.log(pk)
            self._pk_nl[name] = Pk2D(a_arr=a, lk_arr=lk, pk_arr=pk,
                                     is_logp=use_log)

    def _apply_nonlinear_model(self, nonlin_model):
        if nonlin_model is None:
            return

        if not isinstance(nonlin_model, (str, dict)):
            raise ValueError("`nonlin_model` must be str, dict, or None.")
        if isinstance(nonlin_model, str) and not self.has_linear_power:
            raise ValueError("Linear power spectrum is empty.")

        if isinstance(nonlin_model, str):
            nonlin_model = {name: nonlin_model for name in self._pk_lin}

        for name, model in nonlin_model.items():
            if name in self._pk_nl:
                continue

            if name not in self._pk_lin:
                raise KeyError(f"Linear power spectrum {name} does not exist.")
            if model not in [None, "halofit"]:
                raise KeyError(f"{model} is not a valid non-linear model.")
            if name == DEFAULT_POWER_SPECTRUM and model is None:
                raise ValueError("The non-linear matter power spectrum must "
                                 "not be None.")

            if model == 'halofit':
                pkl = self._pk_lin[name]
                self._pk_nl[name] = pkl.apply_halofit(self)



__all__ = ("ConcentrationPrada12",)

import numpy as np

from . import Concentration


class ConcentrationPrada12(Concentration):
    """Concentration-mass relation by `Prada et al. 2012
    <https://arXiv.org/abs/1104.5130>`_. This parametrization is only valid for
    S.O. masses with :math:`\\Delta = 200` times the critical density.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`): a mass
            definition object ora name string.
    """
    name = 'Prada12'

    def __init__(self, *, mass_def="200c"):
        super().__init__(mass_def=mass_def)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.name != "200c"

    def _setup(self):
        self.c0 = 3.681
        self.c1 = 5.033
        self.al = 6.948
        self.x0 = 0.424
        self.i0 = 1.047
        self.i1 = 1.646
        self.be = 7.386
        self.x1 = 0.526
        self.cnorm = 1. / self._cmin(1.393)
        self.inorm = 1. / self._imin(1.393)

    def _form(self, x, x0, v0, v1, v2):
        # form factor for `cmin` and `imin`
        return v0 + (v1 - v0) * (np.arctan(v2 * (x - x0)) / np.pi + 0.5)

    def _cmin(self, x):
        return self._form(x, x0=self.x0, v0=self.c0, v1=self.c1, v2=self.al)

    def _imin(self, x):
        return self._form(x, x0=self.x1, v0=self.i0, v1=self.i1, v2=self.be)

    def _concentration(self, cosmo, M, a):
        sig = cosmo.sigmaM(M, a)
        x = a * (cosmo["Omega_l"] / cosmo["Omega_m"])**(1. / 3.)
        B0 = self._cmin(x) * self.cnorm
        B1 = self._imin(x) * self.inorm
        sig_p = B1 * sig
        Cc = 2.881 * ((sig_p / 1.257)**1.022 + 1) * np.exp(0.060 / sig_p**2)
        return B0 * Cc



import numpy as np
import pyccl as ccl
import pytest


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
TRS = {'TG': ccl.nl_pt.PTNumberCountsTracer(b1=2.0, b2=2.0, bs=2.0, bk2=2.0),
       'TM': ccl.nl_pt.PTMatterTracer()}
PTC = ccl.nl_pt.LagrangianPTCalculator(cosmo=COSMO)


def test_lpt_calculator_smoke():
    c = ccl.nl_pt.LagrangianPTCalculator(log10k_min=-3,
                                         log10k_max=1,
                                         nk_per_decade=10)
    assert len(c.k_s) == 40


@pytest.mark.parametrize('tr1,tr2',
                         [['TG', 'TG'],
                          ['TG', 'TM'],
                          ['TM', 'TG'],
                          ['TM', 'TM']])
def test_lpt_get_pk2d_smoke(tr1, tr2):
    t2 = None if tr2 == tr1 else TRS[tr2]
    ptc = ccl.nl_pt.LagrangianPTCalculator(cosmo=COSMO)

    pk = ptc.get_biased_pk2d(TRS[tr1], tracer2=t2)
    assert isinstance(pk, ccl.Pk2D)


@pytest.mark.parametrize('nl', ['nonlinear', 'linear', 'pt'])
def test_lpt_get_pk2d_nl(nl):
    ptc = ccl.nl_pt.LagrangianPTCalculator(
        b1_pk_kind=nl, bk2_pk_kind=nl, cosmo=COSMO)
    pk = ptc.get_biased_pk2d(TRS['TG'])
    assert isinstance(pk, ccl.Pk2D)


@pytest.mark.parametrize('typ_nlin,typ_nloc', [('linear', 'nonlinear'),
                                               ('nonlinear', 'linear')])
def test_lpt_k2pk_types(typ_nlin, typ_nloc):
    tg = ccl.nl_pt.PTNumberCountsTracer(1., 0., 0., bk2=1.)
    tm = ccl.nl_pt.PTNumberCountsTracer(1., 0., 0.)
    ptc1 = ccl.nl_pt.LagrangianPTCalculator(
        b1_pk_kind=typ_nlin, bk2_pk_kind=typ_nloc, cosmo=COSMO)
    ptc2 = ccl.nl_pt.LagrangianPTCalculator(
        b1_pk_kind=typ_nloc, cosmo=COSMO)
    pkmm = ptc1.get_biased_pk2d(tm, tracer2=tm)
    pkmm2 = ptc2.get_biased_pk2d(tm, tracer2=tm)
    pkgg = ptc1.get_biased_pk2d(tg, tracer2=tg)
    ks = np.geomspace(1E-3, 1E1, 128)
    p1 = pkgg(ks, 1., cosmo=COSMO)
    p2 = pkmm(ks, 1., cosmo=COSMO)+ks**2*pkmm2(ks, 1., cosmo=COSMO)
    assert np.allclose(p1, p2, atol=0, rtol=1E-4)


@pytest.mark.parametrize('kind', ['m:m', 'm:b2', 'm:b3nl', 'm:bs',
                                  'b1:b3nl', 'b3nl:b3nl', 'b3nl:bk2'])
def test_lpt_deconstruction(kind):
    ptc = ccl.nl_pt.LagrangianPTCalculator(cosmo=COSMO,
                                           b1_pk_kind='nonlinear',
                                           bk2_pk_kind='nonlinear')
    b_nc = ['b1', 'b2', 'b3nl', 'bs', 'bk2']
    pk1 = ptc.get_pk2d_template(kind)

    def get_tr(tn):
        if tn == 'm':
            return ccl.nl_pt.PTMatterTracer()
        if tn in b_nc:
            bdict = {b: 0.0 for b in b_nc}
            bdict[tn] = 1.0
            return ccl.nl_pt.PTNumberCountsTracer(
                b1=bdict['b1'], b2=bdict['b2'],
                bs=bdict['bs'], bk2=bdict['bk2'],
                b3nl=bdict['b3nl'])

    tn1, tn2 = kind.split(':')
    t1 = get_tr(tn1)
    t2 = get_tr(tn2)

    pk2 = ptc.get_biased_pk2d(t1, tracer2=t2)
    if pk1 is None:
        assert pk2(0.5, 1.0, cosmo=COSMO) == 0.0
    else:
        v1 = pk1(0.5, 1.0, cosmo=COSMO)
        v2 = pk2(0.5, 1.0, cosmo=COSMO)
        assert np.allclose(v1, v2, atol=0, rtol=1e-6)
        # Check cached
        pk3 = ptc._pk2d_temp[ccl.nl_pt.lpt._PK_ALIAS[kind]]
        v3 = pk3(0.5, 1.0, cosmo=COSMO)
        assert np.allclose(v1, v3, atol=0, rtol=1e-6)


def test_lpt_pk_cutoff():
    # Tests the exponential cutoff
    ks = np.geomspace(1E-2, 15., 128)

    t = ccl.nl_pt.PTNumberCountsTracer(b1=1.0)
    ptc1 = ccl.nl_pt.LagrangianPTCalculator(cosmo=COSMO)
    pk2d1 = ptc1.get_biased_pk2d(t, tracer2=t)
    pk1 = pk2d1(ks, 1.0, cosmo=COSMO)
    ptc2 = ccl.nl_pt.LagrangianPTCalculator(k_cutoff=10.,
                                            n_exp_cutoff=2.,
                                            cosmo=COSMO)
    pk2d2 = ptc2.get_biased_pk2d(t, tracer2=t)
    pk2 = pk2d2(ks, 1.0, cosmo=COSMO)

    expcut = np.exp(-(ks/10.)**2)
    assert np.allclose(pk1*expcut, pk2, atol=0, rtol=1E-3)


def test_lpt_matter_1loop():
    # Check P(k) for linear tracer with b1=1 is the same
    # as matter P(k)
    ptc = ccl.nl_pt.LagrangianPTCalculator(cosmo=COSMO,
                                           b1_pk_kind='pt')
    tg = ccl.nl_pt.PTNumberCountsTracer(b1=1.0)
    tm = ccl.nl_pt.PTMatterTracer()
    ks = np.geomspace(1E-3, 10, 64)
    pk1 = ptc.get_biased_pk2d(tm)(ks, 1.0, cosmo=COSMO)
    pk2 = ptc.get_biased_pk2d(tg)(ks, 1.0, cosmo=COSMO)
    assert np.allclose(pk1, pk2, atol=0, rtol=1E-3)


def test_lpt_calculator_raises():
    # Wrong type of b1 and bk2 power spectra
    with pytest.raises(ValueError):
        ccl.nl_pt.LagrangianPTCalculator(b1_pk_kind='non-linear')

    # Wrong type of b1 and bk2 power spectra
    with pytest.raises(ValueError):
        ccl.nl_pt.LagrangianPTCalculator(bk2_pk_kind='non-linear')

    # Uninitialized templates
    with pytest.raises(ccl.CCLError):
        ptc = ccl.nl_pt.LagrangianPTCalculator()
        ptc.get_biased_pk2d(TRS['TG'])

    # TODO: Discuss this test
    # Wrong pair combination
    with pytest.raises(ValueError):
        ptc = ccl.nl_pt.LagrangianPTCalculator(cosmo=COSMO)
        ptc.get_pk2d_template('b1:b3')


def test_lpt_template_swap():
    # Test that swapping operator order gets you the same
    # Pk
    ks = np.array([0.01, 0.1, 1.0])
    ptc = ccl.nl_pt.LagrangianPTCalculator(cosmo=COSMO)
    pk1 = ptc.get_pk2d_template('b2:bs')(ks, 1.0, cosmo=COSMO)
    pk2 = ptc.get_pk2d_template('bs:b2')(ks, 1.0, cosmo=COSMO)
    assert np.all(pk1 == pk2)


def test_lpt_eq():
    ptc1 = ccl.nl_pt.LagrangianPTCalculator()
    # Should be the same
    ptc2 = ccl.nl_pt.LagrangianPTCalculator()
    assert ptc1 == ptc2
    # Should still be the same
    ptc2 = ccl.nl_pt.LagrangianPTCalculator(
        a_arr=ccl.pyutils.get_pk_spline_a())
    assert ptc1 == ptc2
    # Different a sampling
    ptc2 = ccl.nl_pt.LagrangianPTCalculator(
        a_arr=np.linspace(0.5, 1., 30))
    assert ptc1 != ptc2



import copy
import numpy as np
import pytest
import pyccl
from pyccl import ccllib, CCLError

pyccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
PYCOSMO = pyccl.Cosmology(Omega_c=0.27, Omega_b=0.045, h=0.67,
                          sigma8=0.8, n_s=0.96,
                          transfer_function='bbks')
PYCOSMO.compute_nonlin_power()
COSMO = PYCOSMO.cosmo


def test_swig_tracer():
    z = np.linspace(0, 1, 10)
    bias_ia = z * 2
    dNdz = z * 8

    with pytest.raises(ValueError):
        pyccl.WeakLensingTracer(PYCOSMO, dndz=(z, dNdz[0:2]),
                                ia_bias=(z, bias_ia))

    with pytest.raises(ValueError):
        pyccl.WeakLensingTracer(PYCOSMO, dndz=(z, dNdz),
                                ia_bias=(z, bias_ia[0:2]))


def test_swig_background():
    status = 0
    for func in [
            ccllib.growth_factor_vec,
            ccllib.growth_factor_unnorm_vec,
            ccllib.growth_rate_vec,
            ccllib.comoving_radial_distance_vec,
            ccllib.comoving_angular_distance_vec,
            ccllib.h_over_h0_vec,
            ccllib.luminosity_distance_vec,
            ccllib.distance_modulus_vec]:
        with pytest.raises(CCLError):
            func(COSMO, [0.0, 1.0], 1, status)

    with pytest.raises(CCLError):
        ccllib.omega_x_vec(COSMO, 0, [0.0, 1.0], 1, status)

    with pytest.raises(CCLError):
        ccllib.rho_x_vec(COSMO, 0, 0, [0.0, 1.0], 1, status)

    with pytest.raises(CCLError):
        ccllib.scale_factor_of_chi_vec(COSMO, [0.0, 1.0], 1, status)


def test_swig_cls():
    status = 0
    base_args = [
        [0.0, 1.0],
        [0.0, 1.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0],
        [0.0],
        [0.0, 1.0, 2.0],
        [0.0, 1.0, 2.0, 3.0],
        0, 1, 0, 0, 0, 0, 2,
        status]
    args = copy.deepcopy(base_args)
    args[1] = [1.0] * 8
    with pytest.raises(CCLError):
        ccllib.cl_tracer_t_new_wrapper(COSMO, 0, 0, *args)

    with pytest.raises(CCLError):
        ccllib.angular_cl_vec(COSMO, None, None, None, 1, 0,
                              pyccl.pyutils.integ_types['spline'],
                              "none", status)


def test_swig_correlation():
    status = 0
    with pytest.raises(CCLError):
        ccllib.correlation_vec(
            COSMO,
            [1, 2],
            [1, 2, 3],
            [0, 1, 2, 3],
            0, 0,
            4,
            status)

    with pytest.raises(CCLError):
        ccllib.correlation_vec(
            COSMO,
            [1, 2, 3],
            [1, 2, 3],
            [0, 1, 2, 3],
            0, 0,
            5,
            status)

    with pytest.raises(CCLError):
        ccllib.correlation_3d_vec(
            COSMO,
            PYCOSMO._pk_nl['delta_matter:delta_matter'].psp,
            1.0,
            [1, 2, 3],
            4,
            status)


def test_swig_neurtinos():
    status = 0
    with pytest.raises(CCLError):
        ccllib.Omeganuh2_vec(
            3, 2.7, 0.72,
            [0.0, 1.0],
            [0.05, 0.1, 0.2],
            4,
            status)

    with pytest.raises(CCLError):
        ccllib.Omeganuh2_vec(
            3, 2.7, 0.72,
            [0.0, 1.0],
            [0.1, 0.2],
            2,
            status)


def test_swig_power():
    status = 0

    for func in [ccllib.sigmaR_vec, ccllib.sigmaV_vec]:
        with pytest.raises(CCLError):
            func(COSMO, None, 1.0, [1.0, 2.0], 3, status)

    with pytest.raises(CCLError):
        ccllib.kNL_vec(COSMO, None, [0.5, 1.0], 3, status)


pyccl.gsl_params.reload()  # reset to the default parameters



"""Utility functions to analyze status and error messages passed from CCL, as
well as wrappers to automatically vectorize functions.
"""
__all__ = (
    "CLevelErrors", "ExtrapolationMethods", "IntegrationMethods", "check",
    "debug_mode", "get_pk_spline_lk", "get_pk_spline_a", "resample_array")

from enum import Enum
from typing import Iterable

import numpy as np

from . import CCLError, lib, spline_params


NoneArr = np.array([])


class IntegrationMethods(Enum):
    QAG_QUAD = "qag_quad"
    SPLINE = "spline"


class ExtrapolationMethods(Enum):
    NONE = "none"
    CONSTANT = "constant"
    LINX_LINY = "linx_liny"
    LINX_LOGY = "linx_logy"
    LOGX_LINY = "logx_liny"
    LOGX_LOGY = "logx_logy"


integ_types = {
    'qag_quad': lib.integration_qag_quad,
    'spline': lib.integration_spline}

extrap_types = {
    'none': lib.f1d_extrap_0,
    'constant': lib.f1d_extrap_const,
    'linx_liny': lib.f1d_extrap_linx_liny,
    'linx_logy': lib.f1d_extrap_linx_logy,
    'logx_liny': lib.f1d_extrap_logx_liny,
    'logx_logy': lib.f1d_extrap_logx_logy}

# This is defined here instead of in `errors.py` because SWIG needs `CCLError`
# from `.errors`, resulting in a cyclic import.
CLevelErrors = {
    lib.CCL_ERROR_CLASS: 'CCL_ERROR_CLASS',
    lib.CCL_ERROR_INCONSISTENT: 'CCL_ERROR_INCONSISTENT',
    lib.CCL_ERROR_INTEG: 'CCL_ERROR_INTEG',
    lib.CCL_ERROR_LINSPACE: 'CCL_ERROR_LINSPACE',
    lib.CCL_ERROR_MEMORY: 'CCL_ERROR_MEMORY',
    lib.CCL_ERROR_ROOT: 'CCL_ERROR_ROOT',
    lib.CCL_ERROR_SPLINE: 'CCL_ERROR_SPLINE',
    lib.CCL_ERROR_SPLINE_EV: 'CCL_ERROR_SPLINE_EV',
    lib.CCL_ERROR_COMPUTECHI: 'CCL_ERROR_COMPUTECHI',
    lib.CCL_ERROR_MF: 'CCL_ERROR_MF',
    lib.CCL_ERROR_HMF_INTERP: 'CCL_ERROR_HMF_INTERP',
    lib.CCL_ERROR_PARAMETERS: 'CCL_ERROR_PARAMETERS',
    lib.CCL_ERROR_NU_INT: 'CCL_ERROR_NU_INT',
    lib.CCL_ERROR_EMULATOR_BOUND: 'CCL_ERROR_EMULATOR_BOUND',
    lib.CCL_ERROR_MISSING_CONFIG_FILE: 'CCL_ERROR_MISSING_CONFIG_FILE',
}


def check(status, cosmo=None):
    """Check the status returned by a ccllib function.

    Args:
        status (:obj:`int` or :obj:`~pyccl.core.error_types`):
            Flag or error describing the success of a function.
        cosmo (:class:`~pyccl.cosmology.Cosmology`):
            A Cosmology object.

    :meta private:
    """
    # Check for normal status (no action required)
    if status == 0:
        return

    # Get status message from Cosmology object, if there is one
    if cosmo is not None:
        msg = cosmo.cosmo.status_message
    else:
        msg = ""

    # Check for known error status
    if status in CLevelErrors.keys():
        raise CCLError(f"Error {CLevelErrors[status]}: {msg}")

    # Check for unknown error
    if status != 0:
        raise CCLError(f"Error {status}: {msg}")


def debug_mode(debug):
    """Toggle debug mode on or off. If debug mode is on, the C backend is
    forced to print error messages as soon as they are raised, even if the
    flow of the program continues. This makes it easier to track down errors.

    If debug mode is off, the C code will not print errors, and the Python
    wrapper will raise the last error that was detected. If multiple errors
    were raised, all but the last will be overwritten within the C code, so the
    user will not necessarily be informed of the root cause of the error.

    Args:
        debug (:obj:`bool`): Switch debug mode on (True) or off (False).

    """
    if debug:
        lib.set_debug_policy(lib.CCL_DEBUG_MODE_ON)
    else:
        lib.set_debug_policy(lib.CCL_DEBUG_MODE_OFF)


# This function is not used anymore so we don't want Coveralls to
# include it, but we keep it in case it is needed at some point.
# def _vectorize_fn_simple(fn, fn_vec, x,
#                          returns_status=True):  # pragma: no cover
#     """Generic wrapper to allow vectorized (1D array) access to CCL
#     functions with one vector argument (but no dependence on cosmology).
#
#     Args:
#         fn (callable): Function with a single argument.
#         fn_vec (callable): Function that has a vectorized implementation in
#                            a .i file.
#         x (:obj:`float` or `array`): Argument to fn.
#         returns_stats (bool): Indicates whether fn returns a status.
#
#     """
#     status = 0
#     if isinstance(x, int):
#         x = float(x)
#     if isinstance(x, float):
#         # Use single-value function
#         if returns_status:
#             f, status = fn(x, status)
#         else:
#             f = fn(x)
#     elif isinstance(x, np.ndarray):
#         # Use vectorised function
#         if returns_status:
#             f, status = fn_vec(x, x.size, status)
#         else:
#             f = fn_vec(x, x.size)
#     else:
#         # Use vectorised function
#         if returns_status:
#             f, status = fn_vec(x, len(x), status)
#         else:
#             f = fn_vec(x, len(x))
#
#     # Check result and return
#     check(status)
#     return f


def _vectorize_fn(fn, fn_vec, cosmo, x, returns_status=True):
    """Generic wrapper to allow vectorized (1D array) access to CCL
    functions with one vector argument, with a cosmology dependence.

    Args:
        fn (callable): Function with a single argument.
        fn_vec (callable): Function that has a vectorized implementation in
                           a .i file.
        cosmo (ccl_cosmology or Cosmology): The input cosmology which gets
                                            converted to a ccl_cosmology.
        x (:obj:`float` or `array`): Argument to fn.
        returns_stats (bool): Indicates whether fn returns a status.

    """

    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    status = 0

    if isinstance(x, int):
        x = float(x)
    if isinstance(x, float):
        # Use single-value function
        if returns_status:
            f, status = fn(cosmo, x, status)
        else:
            f = fn(cosmo, x)
    elif isinstance(x, np.ndarray):
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, x, x.size, status)
        else:
            f = fn_vec(cosmo, x, x.size)
    else:
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, x, len(x), status)
        else:
            f = fn_vec(cosmo, x, len(x))

    # Check result and return
    check(status, cosmo_in)
    return f


def _vectorize_fn3(fn, fn_vec, cosmo, x, n, returns_status=True):
    """Generic wrapper to allow vectorized (1D array) access to CCL
    functions with one vector argument and one integer argument,
    with a cosmology dependence.

    Args:
        fn (callable): Function with a single argument.
        fn_vec (callable): Function that has a vectorized implementation in
                           a .i file.
        cosmo (ccl_cosmology or Cosmology): The input cosmology which gets
                                            converted to a ccl_cosmology.
        x (:obj:`float` or `array`): Argument to fn.
        n (:obj:`int`): Integer argument to fn.
        returns_stats (bool): Indicates whether fn returns a status.

    """
    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo
    status = 0

    if isinstance(x, int):
        x = float(x)
    if isinstance(x, float):
        # Use single-value function
        if returns_status:
            f, status = fn(cosmo, x, n, status)
        else:
            f = fn(cosmo, x, n)
    elif isinstance(x, np.ndarray):
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, n, x, x.size, status)
        else:
            f = fn_vec(cosmo, n, x, x.size)
    else:
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, n, x, len(x), status)
        else:
            f = fn_vec(cosmo, n, x, len(x))

    # Check result and return
    check(status, cosmo_in)
    return f


def _vectorize_fn4(fn, fn_vec, cosmo, x, a, d, returns_status=True):
    """Generic wrapper to allow vectorized (1D array) access to CCL
    functions with one vector argument and two float arguments, with
    a cosmology dependence.

    Args:
        fn (callable): Function with a single argument.
        fn_vec (callable): Function that has a vectorized implementation in
                           a .i file.
        cosmo (ccl_cosmology or Cosmology): The input cosmology which gets
                                            converted to a ccl_cosmology.
        x (:obj:`float` or `array`): Argument to fn.
        a (float): Float argument to fn.
        d (float): Float argument to fn.
        returns_stats (bool): Indicates whether fn returns a status.

    """
    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo
    status = 0

    if isinstance(x, int):
        x = float(x)
    if isinstance(x, float):
        if returns_status:
            f, status = fn(cosmo, x, a, d, status)
        else:
            f = fn(cosmo, x, a, d)
    elif isinstance(x, np.ndarray):
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, a, d, x, x.size, status)
        else:
            f = fn_vec(cosmo, a, d, x, x.size)
    else:
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, a, d, x, len(x), status)
        else:
            f = fn_vec(cosmo, a, d, x, len(x))

    # Check result and return
    check(status, cosmo_in)
    return f


def _vectorize_fn5(fn, fn_vec, cosmo, x1, x2, returns_status=True):
    """Generic wrapper to allow vectorized (1D array) access to CCL
    functions with two vector arguments of the same length,
    with a cosmology dependence.

    Args:
        fn (callable): Function with a single argument.
        fn_vec (callable): Function that has a vectorized implementation in
                           a .i file.
        cosmo (ccl_cosmology or Cosmology): The input cosmology which gets
                                            converted to a ccl_cosmology.
        x1 (:obj:`float` or `array`): Argument to fn.
        x2 (:obj:`float` or `array`): Argument to fn.
        returns_stats (bool): Indicates whether fn returns a status.

    """
    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo
    status = 0

    # If a scalar was passed, convert to an array
    if isinstance(x1, int):
        x1 = float(x1)
        x2 = float(x2)
    if isinstance(x1, float):
        # Use single-value function
        if returns_status:
            f, status = fn(cosmo, x1, x2, status)
        else:
            f = fn(cosmo, x1, x2)
    elif isinstance(x1, np.ndarray):
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, x1, x2, x1.size, status)
        else:
            f = fn_vec(cosmo, x1, x2, x1.size)
    else:
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, x1, x2, len(x2), status)
        else:
            f = fn_vec(cosmo, x1, x2, len(x2))

    # Check result and return
    check(status, cosmo_in)
    return f


def _vectorize_fn6(fn, fn_vec, cosmo, x1, x2, returns_status=True):
    """Generic wrapper to allow vectorized (1D array) access to CCL
    functions with two vector arguments of the any length,
    with a cosmology dependence.

    Args:
        fn (callable): Function with a single argument.
        fn_vec (callable): Function that has a vectorized implementation in
                           a .i file.
        cosmo (ccl_cosmology or Cosmology): The input cosmology which gets
                                            converted to a ccl_cosmology.
        x1 (:obj:`float` or `array`): Argument to fn.
        x2 (:obj:`float` or `array`): Argument to fn.
        returns_stats (bool): Indicates whether fn returns a status.

    """
    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo
    status = 0

    # If a scalar was passed, convert to an array
    if isinstance(x1, int):
        x1 = float(x1)
        x2 = float(x2)
    if isinstance(x1, float):
        # Use single-value function
        if returns_status:
            f, status = fn(cosmo, x1, x2, status)
        else:
            f = fn(cosmo, x1, x2)
    elif isinstance(x1, np.ndarray):
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, x1, x2, int(x1.size*x2.size), status)
        else:
            f = fn_vec(cosmo, x1, x2, int(x1.size*x2.size))
    else:
        # Use vectorised function
        if returns_status:
            f, status = fn_vec(cosmo, x1, x2, int(len(x1)*len(x2)), status)
        else:
            f = fn_vec(cosmo, x1, x2, int(len(x1)*len(x2)))

    # Check result and return
    check(status, cosmo_in)
    return f


def get_pk_spline_nk(cosmo=None, spline_params=spline_params):
    """Get the number of sampling points in the wavenumber dimension.

    Arguments:
        cosmo (:obj:`~pyccl.ccllib.cosmology` via SWIG):
            Input cosmology.

    :meta private:
    """
    if cosmo is not None:
        return lib.get_pk_spline_nk(cosmo.cosmo)
    ndecades = np.log10(spline_params.K_MAX / spline_params.K_MIN)
    return int(np.ceil(ndecades*spline_params.N_K))


def get_pk_spline_na(cosmo=None, spline_params=spline_params):
    """Get the number of sampling points in the scale factor dimension.

    Arguments:
        cosmo (:obj:`~pyccl.ccllib.cosmology` via SWIG):
            Input cosmology.

    :meta private:
    """
    if cosmo is not None:
        return lib.get_pk_spline_na(cosmo.cosmo)
    return spline_params.A_SPLINE_NA_PK + spline_params.A_SPLINE_NLOG_PK - 1


def get_pk_spline_lk(cosmo=None, spline_params=spline_params):
    """Get a log(k)-array with sampling rate defined by ``ccl.spline_params``
    or by the spline parameters of the input ``cosmo``.

    Arguments:
        cosmo (:obj:`~pyccl.ccllib.cosmology` via SWIG):
            Input cosmology.

    :meta private:
    """
    nk = get_pk_spline_nk(cosmo=cosmo, spline_params=spline_params)
    if cosmo is not None:
        lk_arr, status = lib.get_pk_spline_lk(cosmo.cosmo, nk, 0)
        check(status, cosmo)
        return lk_arr
    lk_arr, status = lib.get_pk_spline_lk_from_params(spline_params, nk, 0)
    check(status)
    return lk_arr


def get_pk_spline_a(cosmo=None, spline_params=spline_params):
    """Get an a-array with sampling rate defined by ``ccl.spline_params``
    or by the spline parameters of the input ``cosmo``.

    Arguments:
        cosmo (:obj:`~pyccl.ccllib.cosmology` via SWIG):
            Input cosmology.

    :meta private:
    """
    na = get_pk_spline_na(cosmo=cosmo, spline_params=spline_params)
    if cosmo is not None:
        a_arr, status = lib.get_pk_spline_a(cosmo.cosmo, na, 0)
        check(status, cosmo)
        return a_arr
    a_arr, status = lib.get_pk_spline_a_from_params(spline_params, na, 0)
    check(status)
    return a_arr


def resample_array(x_in, y_in, x_out,
                   extrap_lo='none', extrap_hi='none',
                   fill_value_lo=0, fill_value_hi=0):
    """ Interpolates an input y array onto a set of x values.

    Args:
        x_in (`array`): input x-values.
        y_in (`array`): input y-values.
        x_out (`array`): x-values for output array.
        extrap_lo (:obj:`str`): type of extrapolation for x-values below the
            range of `x_in`. 'none' (for no interpolation), 'constant',
            'linx_liny' (linear in x and y), 'linx_logy', 'logx_liny' and
            'logx_logy'.
        extrap_hi (:obj:`str`): type of extrapolation for x-values above the
            range of `x_in`.
        fill_value_lo (:obj:`float`): constant value if `extrap_lo` is
            'constant'.
        fill_value_hi (:obj:`float`): constant value if `extrap_hi` is
            'constant'.
    Returns:
        `array`: output array.
    """
    # TODO: point to the enum in CCLv3 docs.
    if extrap_lo not in extrap_types.keys():
        raise ValueError("Invalid extrapolation type.")
    if extrap_hi not in extrap_types.keys():
        raise ValueError("Invalid extrapolation type.")

    status = 0
    y_out, status = lib.array_1d_resample(x_in, y_in, x_out,
                                          fill_value_lo, fill_value_hi,
                                          extrap_types[extrap_lo],
                                          extrap_types[extrap_hi],
                                          x_out.size, status)
    check(status)
    return y_out


def _fftlog_transform(rs, frs,
                      dim, mu, power_law_index):
    if np.ndim(rs) != 1:
        raise ValueError("rs should be a 1D array")
    if np.ndim(frs) < 1 or np.ndim(frs) > 2:
        raise ValueError("frs should be a 1D or 2D array")
    if np.ndim(frs) == 1:
        n_transforms = 1
        n_r = len(frs)
    else:
        n_transforms, n_r = frs.shape

    if len(rs) != n_r:
        raise ValueError(f"rs should have {n_r} elements")

    status = 0
    result, status = lib.fftlog_transform(n_transforms,
                                          rs, frs.flatten(),
                                          dim, mu, power_law_index,
                                          (n_transforms + 1) * n_r,
                                          status)
    check(status)
    result = result.reshape([n_transforms + 1, n_r])
    ks = result[0]
    fks = result[1:]
    if np.ndim(frs) == 1:
        fks = fks.squeeze()

    return ks, fks


def _spline_integrate(x, ys, a, b):
    if np.ndim(x) != 1:
        raise ValueError("x should be a 1D array")
    if np.ndim(ys) < 1 or np.ndim(ys) > 2:
        raise ValueError("ys should be 1D or a 2D array")
    if np.ndim(ys) == 1:
        n_integ = 1
        n_x = len(ys)
    else:
        n_integ, n_x = ys.shape

    if len(x) != n_x:
        raise ValueError(f"x should have {n_x} elements")

    if np.ndim(a) > 0 or np.ndim(b) > 0:
        raise TypeError("Integration limits should be scalar")

    status = 0
    result, status = lib.spline_integrate(n_integ,
                                          x, ys.flatten(),
                                          a, b, n_integ,
                                          status)
    check(status)

    if np.ndim(ys) == 1:
        result = result[0]

    return result


def _check_array_params(f_arg, name=None, arr3=False):
    """Check whether an argument `f_arg` passed into the constructor of
    Tracer() is valid.

    If the argument is set to `None`, it will be replaced with a special array
    that signals to the CCL wrapper that this argument is NULL.
    """
    if f_arg is None:
        # Return empty array if argument is None
        f1 = NoneArr
        f2 = NoneArr
        f3 = NoneArr
    else:
        if ((not isinstance(f_arg, Iterable))
            or (len(f_arg) != (3 if arr3 else 2))
            or (not (isinstance(f_arg[0], Iterable)
                     and isinstance(f_arg[1], Iterable)))):
            raise ValueError(f"{name} must be a tuple of two arrays.")

        f1 = np.atleast_1d(np.array(f_arg[0], dtype=float))
        f2 = np.atleast_1d(np.array(f_arg[1], dtype=float))
        if arr3:
            f3 = np.atleast_1d(np.array(f_arg[2], dtype=float))
    if arr3:
        return f1, f2, f3
    else:
        return f1, f2


def _get_spline1d_arrays(gsl_spline):
    """Get array data from a 1D GSL spline.

    Args:
        gsl_spline: `SWIGObject` of gsl_spline
            The SWIG object of the GSL spline.

    Returns:
        xarr: `array`
            The x array of the spline.
        yarr: `array`
            The y array of the spline.
    """
    status = 0
    size, status = lib.get_spline1d_array_size(gsl_spline, status)
    check(status)

    xarr, yarr, status = lib.get_spline1d_arrays(gsl_spline, size, size,
                                                 status)
    check(status)

    return xarr, yarr


def _get_spline2d_arrays(gsl_spline):
    """Get array data from a 2D GSL spline.

    Args:
        gsl_spline: `SWIGObject` of gsl_spline2d *
            The SWIG object of the 2D GSL spline.

    Returns:
        yarr: `array`
            The y array of the spline.
        xarr: `array`
            The x array of the spline.
        zarr: `array`
            The z array of the spline. The shape is (yarr.size, xarr.size).
    """
    status = 0
    x_size, y_size, status = lib.get_spline2d_array_sizes(gsl_spline, status)
    check(status)

    z_size = x_size*y_size
    xarr, yarr, zarr, status = lib.get_spline2d_arrays(gsl_spline,
                                                       x_size, y_size, z_size,
                                                       status)
    check(status)

    return yarr, xarr, zarr.reshape(y_size, x_size)


def _get_spline3d_arrays(gsl_spline, length):
    """Get array data from an array of 2D GSL splines.

    Args:
        gsl_spline (`SWIGObject` of gsl_spline2d **):
            The SWIG object of the 2D GSL spline.
        length (:obj:`int`):
            The length of the 3rd dimension.

    Returns:
        xarr: `array`
            The x array of the spline.
        yarr: `array`
            The y array of the spline.
        zarr: `array`
            The z array of the spline. The shape is (yarr.size, xarr.size).
    """
    status = 0
    x_size, y_size, status = lib.get_spline3d_array_sizes(gsl_spline, status)
    check(status)

    z_size = x_size*y_size*length
    xarr, yarr, zarr, status = lib.get_spline3d_arrays(gsl_spline,
                                                       x_size, y_size, z_size,
                                                       length, status)
    check(status)

    return xarr, yarr, zarr.reshape((length, x_size, y_size))


def check_openmp_version():
    """Return the OpenMP specification release date.
    Return 0 if OpenMP is not working.
    """

    return lib.openmp_version()


def check_openmp_threads():
    """Returns the number of processors available to the device.
    Return 0 if OpenMP is not working.
    """

    return lib.openmp_threads()



# flake8: noqa E402
from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version(__name__)
except PackageNotFoundError:
    pass  # not installed
del version, PackageNotFoundError

# Set the environment variable for default config path
from os import environ, path
if environ.get("CLASS_PARAM_DIR") is None:
    environ["CLASS_PARAM_DIR"] = path.dirname(path.abspath(__file__))
del environ, path

# Patch for deprecated alias in Numpy >= 1.20.0 (used in ISiTGR & FAST-PT).
# Deprecation cycle starts in Numpy 1.20 and ends in Numpy 1.24.
from packaging.version import parse
import numpy
numpy.int = int if parse(numpy.__version__) >= parse("1.20.0") else numpy.int
del parse, numpy

from . import ccllib as lib
from .errors import *
from ._core import *
from .pyutils import *

from .background import *
from .power import *

from .tracers import *
from .cells import *
from .correlations import *
from .covariances import *

from .pk2d import *
from .tk3d import *

from .boltzmann import *
from .baryons import *
from .neutrinos import *
from .emulators import *

from . import halos
from . import nl_pt

from .cosmology import *



#!/usr/bin/env python
"""
Generate a set of CLASS power spectra across a set of sample points in 
cosmological parameter space, and compare with CCL.
"""
from param_space import *
import os, sys

# Need to specify directory containing 'class' executable
CLASS_ROOT = None
if len(sys.argv) > 1: CLASS_ROOT = sys.argv[1]
assert CLASS_ROOT is not None, \
    "Must specify CLASS_ROOT (as argument or in source file)."

PREFIX = "std" # Prefix to use for this run
NSAMP = 100 # No. of sample points in parameter space
SEED = 10 # Random seed to use for sampling
ZVALS = np.arange(0., 3., 0.5) # Redshifts to evaluate P(k) at

# Define parameter space to sample over
param_dict = {
    'h':            (0.55, 0.8),
    'Omega_cdm':    (0.15, 0.35),
    'Omega_b':      (0.018, 0.052),
    'A_s':          (1.5e-9, 2.5e-9),
    'n_s':          (0.94, 0.98)
}

# Check that expected output data directories exist
class_datadir = "%s/data/class" % os.path.abspath(".")
ccl_datadir = "%s/data/ccl" % os.path.abspath(".")
if not os.path.exists(class_datadir): os.makedirs(class_datadir)
if not os.path.exists(ccl_datadir): os.makedirs(ccl_datadir)

# Get root filename for CLASS and CCL filenames
root = "%s/%s" % (class_datadir, PREFIX)
ccl_root = "%s/%s" % (ccl_datadir, PREFIX)

# Generate sample points on Latin hypercube
sample_points = generate_latin_hypercube( samples=NSAMP, param_dict=param_dict,
                                          class_root=CLASS_ROOT, seed=SEED )
save_hypercube("%s_params.dat" % root, sample_points)


# Generate CLASS .ini files
print("Writing CLASS linear .ini files")
generate_class_ini(sample_points, root="%s_lin_std" % root, 
                   nonlinear=False, redshifts=ZVALS)
generate_class_ini(sample_points, root="%s_lin_pre" % root, 
                   nonlinear=False, redshifts=ZVALS)

print("Writing CLASS nonlinear .ini files")
generate_class_ini(sample_points, root="%s_nl_std" % root, 
                   nonlinear=True, redshifts=ZVALS)
generate_class_ini(sample_points, root="%s_nl_pre" % root, 
                   nonlinear=True, redshifts=ZVALS)


# Run CLASS on generated .ini files
print("Running CLASS on .ini files")
run_class(fname_pattern="%s_lin_std_?????.ini" % root, 
          class_root=CLASS_ROOT, precision=False)
run_class(fname_pattern="%s_lin_pre_?????.ini" % root, 
          class_root=CLASS_ROOT, precision=True)
run_class(fname_pattern="%s_nl_std_?????.ini" % root, 
          class_root=CLASS_ROOT, precision=False)
run_class(fname_pattern="%s_nl_pre_?????.ini" % root, 
          class_root=CLASS_ROOT, precision=True)


# Run CCL for the same sets of parameters
generate_ccl_pspec(sample_points, ccl_root, 
                   class_data_root="%s_lin_std" % root, 
                   zvals=ZVALS, default_params={'mnu': 0.}, mode='std')

generate_ccl_pspec(sample_points, ccl_root, 
                   class_data_root="%s_lin_pre" % root, 
                   zvals=ZVALS, default_params={'mnu': 0.}, mode='pre')

generate_ccl_pspec(sample_points, ccl_root, 
                   class_data_root="%s_nl_std" % root, 
                   zvals=ZVALS, default_params={'mnu': 0.}, 
                   nonlin=True, mode='std')

generate_ccl_pspec(sample_points, ccl_root, 
                   class_data_root="%s_nl_pre" % root, 
                   zvals=ZVALS, default_params={'mnu': 0.}, 
                   nonlin=True, mode='pre')




__all__ = ("ConcentrationIshiyama21",)

import numpy as np
from scipy.optimize import brentq, root_scalar

from ... import lib
from ... import check
from . import Concentration, get_delta_c


class ConcentrationIshiyama21(Concentration):
    """Concentration-mass relation by `Ishiyama et al. 2021
    <http://arxiv.org/abs/2007.14720>`_. This parametrization is only
    valid for S.O. masses with :math:`\\Delta = \\Delta_{\\rm vir}`, or
    :math:`\\Delta=200` or :math:`500` times the critical density.
    By default it will be initialized for :math:`M_{500c}`.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object or a name string.
        relaxed (:obj:`bool`):
            If ``True``, use concentration for relaxed halos. Otherwise,
            use concentration for all halos. Default: ``False``.
        Vmax (:obj:`bool`):
            If ``True``, use the concentration found with the "Vmax"
            numerical method. Otherwise, use the concentration found with
            profile fitting. Default:  ``False``.
    """
    __repr_attrs__ = __eq_attrs__ = ("mass_def", "relaxed", "Vmax",)
    name = 'Ishiyama21'

    def __init__(self, *, mass_def="500c",
                 relaxed=False, Vmax=False):
        self.relaxed = relaxed
        self.Vmax = Vmax
        super().__init__(mass_def=mass_def)

    def _check_mass_def_strict(self, mass_def):
        is_500Vmax = mass_def.Delta == 500 and self.Vmax
        return mass_def.name not in ["vir", "200c", "500c"] or is_500Vmax

    def _setup(self):
        # key: (Vmax, relaxed, Delta)
        vals = {(True, True, 200): (1.79, 2.15, 2.06, 0.88, 9.24, 0.51),
                (True, False, 200): (1.10, 2.30, 1.64, 1.72, 3.60, 0.32),
                (False, True, 200): (0.60, 2.14, 2.63, 1.69, 6.36, 0.37),
                (False, False, 200): (1.19, 2.54, 1.33, 4.04, 1.21, 0.22),
                (True, True, "vir"): (2.40, 2.27, 1.80, 0.56, 13.24, 0.079),
                (True, False, "vir"): (0.76, 2.34, 1.82, 1.83, 3.52, -0.18),
                (False, True, "vir"): (1.22, 2.52, 1.87, 2.13, 4.19, -0.017),
                (False, False, "vir"): (1.64, 2.67, 1.23, 3.92, 1.30, -0.19),
                (False, True, 500): (0.38, 1.44, 3.41, 2.86, 2.99, 0.42),
                (False, False, 500): (1.83, 1.95, 1.17, 3.57, 0.91, 0.26)}

        key = (self.Vmax, self.relaxed, self.mass_def.Delta)
        self.kappa, self.a0, self.a1, \
            self.b0, self.b1, self.c_alpha = vals[key]

    def _dlsigmaR(self, cosmo, M, a):
        # kappa multiplies radius, so in log, 3*kappa multiplies mass
        logM = 3*np.log10(self.kappa) + np.log10(M)

        status = 0
        dlns_dlogM, status = lib.dlnsigM_dlogM_vec(cosmo.cosmo, a, logM,
                                                   len(logM), status)
        check(status, cosmo=cosmo)
        return -3/np.log(10) * dlns_dlogM

    def _G(self, x, n_eff):
        fx = np.log(1 + x) - x / (1 + x)
        G = x / fx**((5 + n_eff) / 6)
        return G

    def _G_inv(self, arg, n_eff):
        # Numerical calculation of the inverse of `_G`.
        roots = []
        for val, neff in zip(arg, n_eff):
            func = lambda x: self._G(x, neff) - val  # noqa: _G_inv Traceback
            try:
                rt = brentq(func, a=0.05, b=200)
            except ValueError:
                # No root in [0.05, 200] (rare, but it may happen).
                rt = root_scalar(func, x0=1, x1=2).root.item()
            roots.append(rt)
        return np.asarray(roots)

    def _concentration(self, cosmo, M, a):
        nu = get_delta_c(cosmo, a, 'EdS_approx') / cosmo.sigmaM(M, a)
        n_eff = -2 * self._dlsigmaR(cosmo, M, a) - 3
        alpha_eff = cosmo.growth_rate(a)

        A = self.a0 * (1 + self.a1 * (n_eff + 3))
        B = self.b0 * (1 + self.b1 * (n_eff + 3))
        C = 1 - self.c_alpha * (1 - alpha_eff)
        arg = A / nu * (1 + nu**2 / B)
        G = self._G_inv(arg, n_eff)
        return C * G



__all__ = ("ConcentrationBhattacharya13",)

from . import Concentration, get_delta_c


class ConcentrationBhattacharya13(Concentration):
    """Concentration-mass relation by `Bhattacharya et al. 2013
    <https://arxiv.org/abs/1112.5479>`_. This parametrization is only valid for
    S.O. masses with :math:`\\Delta = \\Delta_{\\rm vir}`, or
    :math:`\\Delta=200` times the critical or matter densities. By default it
    will be initialized for :math:`M_{200c}`.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`): a mass
            definition object or name string.
    """
    name = 'Bhattacharya13'

    def __init__(self, *, mass_def="200c"):
        super().__init__(mass_def=mass_def)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.name not in ["vir", "200m", "200c"]

    def _setup(self):
        vals = {"vir": (7.7, 0.9, -0.29),
                "200m": (9.0, 1.15, -0.29),
                "200c": (5.9, 0.54, -0.35)}

        self.A, self.B, self.C = vals[self.mass_def.name]

    def _concentration(self, cosmo, M, a):
        gz = cosmo.growth_factor(a)
        delta_c = get_delta_c(cosmo, a, kind='NakamuraSuto97')
        sig = cosmo.sigmaM(M, a)
        nu = delta_c / sig
        return self.A * gz**self.B * nu**self.C



__all__ = ("HaloBiasSheth99",)

from . import HaloBias, get_delta_c


class HaloBiasSheth99(HaloBias):
    """Implements halo bias as described in `Sheth & Tormen 1999
    <https://arxiv.org/abs/astro-ph/9901122>`_.
    This parametrization is only valid for 'fof' masses.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
        use_delta_c_fit (:obj:`bool`): if ``True``, use the fit to the
            critical overdensity :math:`\\delta_c` by
            `Nakamura & Suto 1997
            <https://arxiv.org/abs/astro-ph/9612074>`_. Otherwise use
            :math:`\\delta_c = 1.68647`.
    """
    __repr_attrs__ = __eq_attrs__ = ("mass_def", "mass_def_strict",
                                     "use_delta_c_fit",)
    name = "Sheth99"

    def __init__(self, *,
                 mass_def="fof",
                 mass_def_strict=True,
                 use_delta_c_fit=False):
        self.use_delta_c_fit = use_delta_c_fit
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.Delta != "fof"

    def _setup(self):
        self.p = 0.3
        self.a = 0.707

    def _get_bsigma(self, cosmo, sigM, a):
        if self.use_delta_c_fit:
            delta_c = get_delta_c(cosmo, a, kind='NakamuraSuto97')
        else:
            delta_c = get_delta_c(cosmo, a, kind='EdS')

        nu = delta_c / sigM
        anu2 = self.a * nu**2
        return 1 + (anu2 - 1. + 2. * self.p / (1. + anu2**self.p))/delta_c



import numpy as np
import pytest
import pyccl as ccl


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='halofit')


@pytest.mark.parametrize('m', [
    1e14,
    int(1e14),
    [1e14, 1e15],
    np.array([1e14, 1e15])])
def test_massfunc_smoke(m):
    a = 0.8
    mf = ccl.halos.MassFuncTinker10()(COSMO, m, a)
    assert np.all(np.isfinite(mf))
    assert np.shape(mf) == np.shape(m)


@pytest.mark.parametrize('m', [
    1e14,
    int(1e14),
    [1e14, 1e15],
    np.array([1e14, 1e15])])
def test_sigmaM_smoke(m):
    a = 0.8
    s = ccl.sigmaM(COSMO, m, a)
    assert np.all(np.isfinite(s))
    assert np.shape(s) == np.shape(m)


def test_deltac():
    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05,
                          h=0.7, n_s=0.96, sigma8=0.8,
                          transfer_function='bbks')
    # Test EdS value
    dca = 3*(12*np.pi)**(2/3)/20
    dcb = ccl.halos.get_delta_c(cosmo, 1.0, kind='EdS')
    assert np.fabs(dcb/dca-1) < 1E-5

    # Test Mead et al. value
    dca = (1.59+0.0314*np.log(0.8))*(1+0.0123*np.log10(0.3))
    dcb = ccl.halos.get_delta_c(cosmo, 1.0, kind='Mead16')
    assert np.fabs(dcb/dca-1) < 1E-5



__all__ = ("halomod_power_spectrum", "halomod_Pk2D",)

import numpy as np

from .. import Pk2D
from . import Profile2pt


def halomod_power_spectrum(cosmo, hmc, k, a, prof, *,
                           prof2=None, prof_2pt=None,
                           p_of_k_a=None,
                           get_1h=True, get_2h=True,
                           smooth_transition=None, suppress_1h=None,
                           extrap_pk=False):
    """ Computes the halo model power spectrum for two
    quantities defined by their respective halo profiles.
    The halo model power spectrum for two profiles
    :math:`u` and :math:`v` is:

    .. math::
        P_{u,v}(k,a) = I^0_2(k,a|u,v) +
        I^1_1(k,a|u)\\,I^1_1(k,a|v)\\,P_{\\rm lin}(k,a)

    where :math:`P_{\\rm lin}(k,a)` is the linear matter
    power spectrum, :math:`I^1_1` is defined in the documentation
    of :meth:`~pyccl.halos.halo_model.HMCalculator.I_1_1`, and :math:`I^0_2`
    is defined in the documentation of
    :meth:`~pyccl.halos.halo_model.HMCalculator.I_0_2`.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        hmc (:class:`~pyccl.halos.halo_model.HMCalculator`): a halo model calculator.
        k (:obj:`float` or `array`): comoving wavenumber in Mpc^-1.
        a (:obj:`float` or `array`): scale factor.
        prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile.
        prof2 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): a
            second halo profile. If ``None``, ``prof`` will be used as
            ``prof2``.
        prof_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
            a profile covariance object
            returning the the two-point moment of the two profiles
            being correlated. If ``None``, the default second moment
            will be used, corresponding to the products of the means
            of both profiles.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`): a `Pk2D` object to
            be used as the linear matter power spectrum. If ``None``,
            the power spectrum stored within `cosmo` will be used.
        get_1h (:obj:`bool`): if ``False``, the 1-halo term (i.e. the first
            term in the first equation above) won't be computed.
        get_2h (:obj:`bool`): if ``False``, the 2-halo term (i.e. the second
            term in the first equation above) won't be computed.
        smooth_transition (:obj:`callable` or :obj:`None`):
            Modify the halo model 1-halo/2-halo transition region
            via a time-dependent function :math:`\\alpha(a)`,
            defined as in `HMCODE-2020 <https://arxiv.org/abs/2009.01858>`_:
            :math:`P(k,a)=
            (P_{1h}^{\\alpha(a)}(k)+P_{2h}^{\\alpha(a)}(k))^{1/\\alpha}`.
            If ``None`` the extra factor is just 1.
        suppress_1h (:obj:`callable` or :obj:`None`):
            Suppress the 1-halo large scale contribution by a
            time- and scale-dependent function :math:`k_*(a)`,
            defined as in `HMCODE-2020 <https://arxiv.org/abs/2009.01858>`_:
            :math:`1/[1+(k_*(a)/k)^4]`.
            If ``None`` the standard 1-halo term is returned with no damping.
        extrap_pk (:obj:`bool`):
            Whether to extrapolate ``p_of_k_a`` in case ``a`` is out of its
            support. If ```False```, and the queried values are out of bounds,
            an error is raised.

    Returns:
        (:obj:`float` or `array`): integral values evaluated at each
        combination of ``k`` and ``a``. The shape of the output will
        be ``(N_a, N_k)`` where ``N_k`` and ``N_a`` are the sizes of
        ``k`` and ``a`` respectively. If ``k`` or ``a`` are scalars, the
        corresponding dimension will be squeezed out on output.
    """ # noqa
    a_use = np.atleast_1d(a).astype(float)
    k_use = np.atleast_1d(k).astype(float)

    # Check inputs
    if smooth_transition is not None:
        if not (get_1h and get_2h):
            raise ValueError("Transition region can only be modified "
                             "when both 1-halo and 2-halo terms are queried.")
    if suppress_1h is not None:
        if not get_1h:
            raise ValueError("Can't suppress the 1-halo term "
                             "when get_1h is False.")

    if prof2 is None:
        prof2 = prof
    if prof_2pt is None:
        prof_2pt = Profile2pt()

    pk2d = cosmo.parse_pk(p_of_k_a)
    extrap = cosmo if extrap_pk else None  # extrapolation rule for pk2d

    na = len(a_use)
    nk = len(k_use)
    out = np.zeros([na, nk])
    for ia, aa in enumerate(a_use):
        # normalizations
        norm1 = prof.get_normalization(cosmo, aa, hmc=hmc)

        if prof2 == prof:
            norm2 = norm1
        else:
            norm2 = prof2.get_normalization(cosmo, aa, hmc=hmc)

        if get_2h:
            # bias factors
            i11_1 = hmc.I_1_1(cosmo, k_use, aa, prof)

            if prof2 == prof:
                i11_2 = i11_1
            else:
                i11_2 = hmc.I_1_1(cosmo, k_use, aa, prof2)

            pk_2h = pk2d(k_use, aa, cosmo=extrap) * i11_1 * i11_2  # 2h term
        else:
            pk_2h = 0

        if get_1h:
            pk_1h = hmc.I_0_2(cosmo, k_use, aa, prof,
                              prof2=prof2, prof_2pt=prof_2pt)  # 1h term

            if suppress_1h is not None:
                # large-scale damping of 1-halo term
                ks = suppress_1h(aa)
                pk_1h *= (k_use / ks)**4 / (1 + (k_use / ks)**4)
        else:
            pk_1h = 0

        # smooth 1h/2h transition region
        if smooth_transition is None:
            out[ia] = (pk_1h + pk_2h) / (norm1 * norm2)
        else:
            alpha = smooth_transition(aa)
            out[ia] = (pk_1h**alpha + pk_2h**alpha)**(1/alpha) / (norm1*norm2)

    if np.ndim(a) == 0:
        out = np.squeeze(out, axis=0)
    if np.ndim(k) == 0:
        out = np.squeeze(out, axis=-1)
    return out


def halomod_Pk2D(cosmo, hmc, prof, *,
                 prof2=None, prof_2pt=None,
                 p_of_k_a=None,
                 get_1h=True, get_2h=True,
                 lk_arr=None, a_arr=None,
                 extrap_order_lok=1, extrap_order_hik=2,
                 smooth_transition=None, suppress_1h=None, extrap_pk=False):
    """ Returns a :class:`~pyccl.pk2d.Pk2D` object containing
    the halo-model power spectrum for two quantities defined by
    their respective halo profiles. See :meth:`halomod_power_spectrum`
    for more details about the actual calculation.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        hmc (:class:`~pyccl.halos.halo_model.HMCalculator`): a halo model calculator.
        prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile.
        prof2 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): a
            second halo profile. If ``None``, ``prof`` will be used as
            ``prof2``.
        prof_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
            a profile covariance object
            returning the the two-point moment of the two profiles
            being correlated. If ``None``, the default second moment
            will be used, corresponding to the products of the means
            of both profiles.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`): a `Pk2D` object to
            be used as the linear matter power spectrum. If ``None``,
            the power spectrum stored within `cosmo` will be used.
        get_1h (:obj:`bool`): if ``False``, the 1-halo term (i.e. the first
            term in the first equation above) won't be computed.
        get_2h (:obj:`bool`): if ``False``, the 2-halo term (i.e. the second
            term in the first equation above) won't be computed.
        a_arr (array): an array holding values of the scale factor
            at which the halo model power spectrum should be
            calculated for interpolation. If ``None``, the internal
            values used by `cosmo` will be used.
        lk_arr (array): an array holding values of the natural
            logarithm of the wavenumber (in units of Mpc^-1) at
            which the halo model power spectrum should be calculated
            for interpolation. If ``None``, the internal values used
            by `cosmo` will be used.
        extrap_order_lok (:obj:`int`): extrapolation order to be used on
            k-values below the minimum of the splines. See
            :class:`~pyccl.pk2d.Pk2D`.
        extrap_order_hik (:obj:`int`): extrapolation order to be used on
            k-values above the maximum of the splines. See
            :class:`~pyccl.pk2d.Pk2D`.
        smooth_transition (:obj:`callable` or :obj:`None`):
            Modify the halo model 1-halo/2-halo transition region
            via a time-dependent function :math:`\\alpha(a)`,
            defined as in `HMCODE-2020 <https://arxiv.org/abs/2009.01858>`_:
            :math:`P(k,a)= (P_{1h}^{\\alpha(a)}(k)+P_{2h}^{\\alpha(a)}(k))^{1/\\alpha}`.
            If ``None`` the extra factor is just 1.
        suppress_1h (:obj:`callable` or :obj:`None`):
            Suppress the 1-halo large scale contribution by a
            time- and scale-dependent function :math:`k_*(a)`,
            defined as in `HMCODE-2020 <https://arxiv.org/abs/2009.01858>`_:
            :math:`1/[1+(k_*(a)/k)^4]`.
            If ``None`` the standard 1-halo term is returned with no damping.
        extrap_pk (:obj:`bool`):
            Whether to extrapolate ``p_of_k_a`` in case ``a`` is out of its
            support. If ```False```, and the queried values are out of bounds,
            an error is raised.

    Returns:
        :class:`~pyccl.pk2d.Pk2D`: halo model power spectrum.
    """ # noqa
    if lk_arr is None:
        lk_arr = cosmo.get_pk_spline_lk()
    if a_arr is None:
        a_arr = cosmo.get_pk_spline_a()

    pk_arr = halomod_power_spectrum(
        cosmo, hmc, np.exp(lk_arr), a_arr,
        prof, prof2=prof2, prof_2pt=prof_2pt, p_of_k_a=p_of_k_a,
        get_1h=get_1h, get_2h=get_2h,
        smooth_transition=smooth_transition, suppress_1h=suppress_1h,
        extrap_pk=extrap_pk)

    return Pk2D(a_arr=a_arr, lk_arr=lk_arr, pk_arr=pk_arr,
                extrap_order_lok=extrap_order_lok,
                extrap_order_hik=extrap_order_hik,
                is_logp=False)



"""
Benchmark for the mass-concentration relations derived
using the Uchuu simulations (arXiv:2007.14720).

With many thanks to Tomoaki Ishiyama for kindly providing
the benchmarks.
"""
import os
import pyccl as ccl
import pytest
import numpy as np

dirdat = os.path.join(os.path.dirname(__file__), "data")

# Planck18 used in Uchuu simulations
COSMO = ccl.Cosmology(Omega_c=0.2589, Omega_b=0.0486, h=0.6774,
                      sigma8=0.8159, n_s=0.9667)
COSMO.compute_sigma()
H100 = COSMO["h"]

Z = np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 5.0, 7.0])
# All parametrizations are accurate at to 5% except (500, Vmax, relaxed)
# for which the model is accurate to 10%. This may be due to low mass
# precision error. We thus clip the masses at 8 < log10(M) < 16.
UCHUU_DATA_SCATTER = 0.035
M_min, M_max = 1e8, 1e16


@pytest.mark.parametrize("pars",
                         [{"Delta": 200, "relaxed": False, "Vmax": False},
                          {"Delta": 200, "relaxed": False, "Vmax": True},
                          {"Delta": 200, "relaxed": True, "Vmax": False},
                          {"Delta": 200, "relaxed": True, "Vmax": True},
                          {"Delta": "vir", "relaxed": False, "Vmax": False},
                          {"Delta": "vir", "relaxed": False, "Vmax": True},
                          {"Delta": "vir", "relaxed": True, "Vmax": False},
                          {"Delta": "vir", "relaxed": True, "Vmax": True},
                          {"Delta": 500, "relaxed": False, "Vmax": False},
                          {"Delta": 500, "relaxed": True, "Vmax": False}])
def test_concentration_Ishiyama21(pars):
    Delta = pars["Delta"]
    key1 = "cvmax" if pars["Vmax"] else "crs"
    key2 = "relaxed" if pars["relaxed"] else "all"
    key0 = f"Uchuu_m{Delta}-c{Delta}.{key1}.{key2}"
    fname = os.path.join(dirdat, key0)
    data = np.loadtxt(fname)
    M = data[:, 0]

    # mass cutoff at CCL integration boundaries
    data = data[(M > M_min/H100) & (M < M_max/H100)]
    M_use = M[(M > M_min/H100) & (M < M_max/H100)]

    hmd = ccl.halos.MassDef(Delta, "critical")

    cm = ccl.halos.ConcentrationIshiyama21(mass_def=hmd,
                                           relaxed=pars["relaxed"],
                                           Vmax=pars["Vmax"])

    for i, zz in enumerate(Z):
        dat = data[:, i+1]
        mod = cm(COSMO, M_use/H100, 1/(1+zz))
        assert np.allclose(mod, dat, rtol=UCHUU_DATA_SCATTER)



import pytest
import numpy as np

import pyccl as ccl


def test_camb_class_consistent_smoke(kwargs=None, pkerr=1e-3):
    kwargs = kwargs or {}
    print('kwargs:', kwargs)

    c_camb = ccl.Cosmology(
        Omega_c=0.25, Omega_b=0.05, h=0.7, n_s=0.95, A_s=2e-9,
        transfer_function='boltzmann_camb', **kwargs)

    c_class = ccl.Cosmology(
        Omega_c=0.25, Omega_b=0.05, h=0.7, n_s=0.95, A_s=2e-9,
        transfer_function='boltzmann_class', **kwargs)

    rel_sigma8 = np.abs(ccl.sigma8(c_camb) / ccl.sigma8(c_class) - 1)

    a = 0.845
    k = np.logspace(-3, 1, 100)
    pk_camb = ccl.linear_matter_power(c_camb, k, a)
    pk_class = ccl.linear_matter_power(c_class, k, a)
    rel_pk = np.max(np.abs(pk_camb / pk_class - 1))

    print('rel err pk:', rel_pk)
    print('rel err sigma8:', rel_sigma8)

    assert rel_sigma8 < 3e-3
    assert rel_pk < pkerr


@pytest.mark.parametrize('kwargs', [
    dict(m_nu=0.5),
    dict(m_nu=[0.2, 0.1, 0.0]),
    dict(m_nu=0.5, Neff=4.046)])
def test_camb_class_consistent_nu(kwargs):
    test_camb_class_consistent_smoke(kwargs=kwargs, pkerr=6e-3)


@pytest.mark.parametrize('kwargs', [
    dict(w0=-0.9, wa=0.0),
    dict(w0=-0.9, wa=-0.1)])
def test_camb_class_consistent_de(kwargs):
    test_camb_class_consistent_smoke(kwargs=kwargs, pkerr=1e-3)


@pytest.mark.parametrize('kwargs', [
    dict(m_nu=[0.2, 0.1, 0.5], w0=-0.9, wa=-0.1)])
def test_camb_class_consistent_de_nu(kwargs):
    test_camb_class_consistent_smoke(kwargs=kwargs, pkerr=8e-3)



import pytest
import pyccl as ccl
import functools


def all_subclasses(cls):
    """Get all subclasses of ``cls``. NOTE: Used in ``conftest.py``."""
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


def test_method_control_raises():
    # All method control subclasses must contain a default implementation.
    with pytest.raises(ValueError):
        class MyMethodControl(ccl._core.schema._CustomMethod, method="name"):
            pass


def test_repr_control():
    # Test custom repr controls.
    cosmo = ccl.CosmologyVanillaLCDM()

    ccl.CustomRepr.disable()
    assert repr(cosmo) == object.__repr__(cosmo)

    ccl.CustomRepr.enable()
    assert repr(cosmo) != object.__repr__(cosmo)


def test_eq_control():
    # Test custom eq controls.
    cosmo = [ccl.CosmologyVanillaLCDM() for _ in range(2)]
    assert id(cosmo[0]) != id(cosmo[1])

    ccl.CustomEq.disable()
    assert cosmo[0] != cosmo[1]

    ccl.CustomEq.enable()
    assert cosmo[0] == cosmo[1]


def check_eq_repr_hash(self, other, *, equal=True):
    # Helper to ensure `__eq__`, `__repr__`, `__hash__` are consistent.
    if equal:
        return (self == other
                and repr(self) == repr(other)
                and hash(self) == hash(other))
    return (self != other
            and repr(self) != repr(other)
            and hash(self) != hash(other))


def test_CCLObject_immutability():
    # These tests check the behavior of immutable objects, i.e. instances
    # of classes where `Funlock` or `unlock_instance` is not used.
    # test `CCLObject` lock
    obj = ccl.CCLObject()
    obj._object_lock.unlock()
    assert "locked=False" in repr(obj._object_lock)
    obj._object_lock.lock()
    assert "locked=True" in repr(obj._object_lock)

    # `update_parameters` not implemented.
    cosmo = ccl.CosmologyVanillaLCDM()
    # with pytest.raises(AttributeError):  # TODO: Uncomment for CCLv3.
    #     cosmo.my_attr = "hello_world"
    with pytest.raises(NotImplementedError):
        cosmo.update_parameters(A_SPLINE_NA=120)

    # `update_parameters` implemented.
    prof = ccl.halos.HaloProfilePressureGNFW(mass_def="200c", mass_bias=0.5)
    # with pytest.raises(AttributeError):  # TODO: Uncomment for CCLv3.
    #     prof.mass_bias = 0.7
    assert prof.mass_bias == 0.5
    prof.update_parameters(mass_bias=0.7)
    assert prof.mass_bias == 0.7


def test_CCLObject_default_behavior():
    # Test that if `__repr__` is not defined the fall back is safe.
    MyType = type("MyType", (ccl.CCLObject,), {"test": 0})
    instances = [MyType() for _ in range(2)]
    assert check_eq_repr_hash(*instances, equal=False)

    # Test that all subclasses of ``CCLAutoRepr`` use Python's default
    # ``repr`` if no ``__repr_attrs__`` has been defined.
    instances = [ccl.CCLAutoRepr() for _ in range(2)]
    assert check_eq_repr_hash(*instances, equal=False)

    MyType = type("MyType", (ccl.CCLAutoRepr,), {"test": 0})
    instances = [MyType() for _ in range(2)]
    assert instances[0] != instances[1]


def test_named_class_raises():
    # Test that an error is raised if `create_instance` gets the wrong type.
    with pytest.raises(AttributeError):
        ccl.halos.MassDef.create_instance(1)


def test_ccl_parameters_abstract():
    # Test that the Parameters base class is abstract and cannot instantiate
    # if `instance` or `factory` are not specified.
    with pytest.raises(TypeError):
        ccl.CCLParameters()
    with pytest.raises(ValueError):
        class MyPars(ccl.CCLParameters):
            pass


# +==========================================================================+
# | The following functions are used by `conftest.py` to check correct setup.|
# +==========================================================================+


def init_decorator(func):
    """Check that all attributes listed in ``__repr_attrs__`` are defined in
    the constructor of all subclasses of ``CCLAutoRepr``.
    NOTE: Used in ``conftest.py``.
    """

    def in_mro(self):
        """Determine if `__repr_attrs__` is defined somewhere in the MRO."""
        # NOTE: This helper function makes sure that an AttributeError is not
        # raised when `super().__init__` is called inside another `__init__`.
        mro = self.__class__.__mro__[1:]  # MRO excluding this class
        for cls in mro:
            if hasattr(cls, "__repr_attrs__"):
                return True
        return False

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        if not hasattr(self, "__repr_attrs__"):
            # If `__repr_attrs__` is not specified, use local repr or inherit.
            return

        flag = [attr for attr in self.__repr_attrs__
                if not (hasattr(self, attr) or in_mro(self))]
        if flag:
            # NOTE: Set the attributes before calling `super`.
            raise AttributeError(f"{self.__class__.__name__}: attribute(s) "
                                 f"{flag} not set in __init__.")

    return wrapper


def test_unlock_instance_errors():
    # Test that unlock_instance gives the correct errors.

    # 1. Developer error
    with pytest.raises(NameError):
        @ccl.unlock_instance(name="hello")
        def func1(item, pk, a0=0, *, a1=None, a2):
            return

    # 2. User error
    @ccl.unlock_instance(name="pk")
    def func2(item, pk, a0=0, *, a1=None, a2):
        return

    with pytest.raises(TypeError):
        func2()

    # 3. Doesn't do anything if instance is not CCLObject.
    with ccl.UnlockInstance(True, mutate=False):
        pass



import numpy as np
import pytest
from pyccl.pyutils import _fftlog_transform


def fk(k, alpha, mu, dim):
    return k**(-alpha)


def fr(r, alpha, mu, dim):
    from scipy.special import gamma
    g1 = gamma(0.5 * (dim - alpha + mu))
    g2 = gamma(0.5 * (alpha + mu))
    den = np.pi**(dim/2.) * 2**alpha
    return g1 / (g2 * den * r**(dim - alpha))


@pytest.mark.parametrize('dim', [2, 3])
@pytest.mark.parametrize('mu', [0, 2])
@pytest.mark.parametrize('alpha', [1.2, 1.5, 1.8])
def test_fftlog_plaw(dim, mu, alpha):
    # The d-D Hankel transform of k^{-alpha} is
    # \Gamma[(d - \alpha + \mu) / 2] /
    # \Gamma[(\alpha + \mu) / 2] /
    # (\pi^{d/2} * 2^\alpha * r^{d-\alpha})

    nk = 1024
    k_arr = np.logspace(-4, 4, nk)
    fk_arr = fk(k_arr, alpha, mu, dim)

    r_arr, fr_arr = _fftlog_transform(k_arr, fk_arr,
                                      dim, mu, -alpha)

    fr_arr_pred = fr(r_arr, alpha, mu, dim)
    res = np.fabs(fr_arr / fr_arr_pred - 1)
    assert np.all(res < 1E-10)


def test_fftlog_shapes():
    nk = 1024
    nt = 4
    k_arr = np.logspace(-4, 4, nk)
    fk_arr = []
    for i in range(nt):
        fk_arr.append(fk(k_arr, 1.5, 0, 2))
    fk_arr = np.array(fk_arr)

    # Scalar rs
    with pytest.raises(ValueError):
        _fftlog_transform(k_arr[0], fk_arr,
                          2, 0, 1.5)

    # Scalar frs
    with pytest.raises(ValueError):
        _fftlog_transform(k_arr, fk_arr[0][0],
                          2, 0, 1.5)

    # Wrong rs
    with pytest.raises(ValueError):
        _fftlog_transform(k_arr[1:], fk_arr,
                          2, 0, 1.5)

    # Single transform
    r, fr = _fftlog_transform(k_arr, fk_arr[0],
                              2, 0, 1.5)
    assert fr.shape == (nk,)

    # Multiple transforms
    r, fr = _fftlog_transform(k_arr, fk_arr,
                              2, 0, 1.5)
    assert fr.shape == (nt, nk)



import numpy as np
import pyccl as ccl


def test_satellite_shear_profile():
    k_arr, gamma_k_CosmoSIS = np.loadtxt(
        './benchmarks/data/satellite_shear_profile_cosmoSIS_dev.txt',
        unpack=True)
    cosmo = ccl.Cosmology(Omega_c=0.278 - 0.0391,
                          Omega_b=0.0391, h=0.7,
                          A_s=1.70150098142e-09, n_s=0.978)

    z_eval = 0.2
    mass_eval = 1e15  # Msun

    cM = ccl.halos.ConcentrationDuffy08(mass_def='200m')

    sat_gamma_HOD_simps = ccl.halos.SatelliteShearHOD(concentration=cM, lmax=6,
                                                      a1h=0.000989,
                                                      mass_def='200m')
    gamma_k_CCL = sat_gamma_HOD_simps._usat_fourier(cosmo, k_arr, mass_eval,
                                                    1 / (1 + z_eval))
    # This benchmark is not accurate enough and only serves as a
    # qualitative test, which is why the accuracy criterion is so low.
    assert np.all(np.fabs(-gamma_k_CCL / gamma_k_CosmoSIS - 1) < 0.8)

    # An in-house benchmark test has been developed, testing the accuracy
    # of the FFTLog method against a highly accurate integration method as
    # described in benchmarks/data/codes/halomod_IA_FFTLog_accuracy.ipynb.
    l_arr, cl_GI_benchmark, cl_II_benchmark = np.loadtxt(
        './benchmarks/data/IA_halomodel_Cell_test.txt',
        delimiter=',', unpack=True)
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.045,
                          h=0.67, sigma8=0.83, n_s=0.96)
    k_arr = np.geomspace(1E-3, 1e2, 128)  # For evaluating
    a_arr = np.linspace(0.1, 1, 16)

    cM = ccl.halos.ConcentrationDuffy08(mass_def="200m")
    nM = ccl.halos.MassFuncTinker08(mass_def="200m")
    bM = ccl.halos.HaloBiasTinker10(mass_def="200m")
    hmc = ccl.halos.HMCalculator(mass_function=nM,
                                 halo_bias=bM,
                                 mass_def="200m", nM=64)

    sat_gamma_HOD = ccl.halos.SatelliteShearHOD(concentration=cM,
                                                mass_def='200m')
    NFW = ccl.halos.HaloProfileNFW(concentration=cM, truncated=True,
                                   fourier_analytic=True, mass_def='200m')

    pk_GI_1h = ccl.halos.halomod_Pk2D(cosmo, hmc, NFW,
                                      prof2=sat_gamma_HOD,
                                      get_2h=False,
                                      lk_arr=np.log(k_arr),
                                      a_arr=a_arr)
    pk_II_1h = ccl.halos.halomod_Pk2D(cosmo, hmc, sat_gamma_HOD,
                                      get_2h=False,
                                      lk_arr=np.log(k_arr),
                                      a_arr=a_arr)

    z_arr = np.linspace(0., 3., 256)
    z0 = 0.1
    pz = 1. / (2. * z0) * (z_arr / z0) ** 2. * np.exp(-z_arr / z0)

    b_IA = np.ones(len(z_arr))  # A_IA = 1 in the NLA model
    ia_tracer = ccl.WeakLensingTracer(cosmo, dndz=(z_arr, pz), has_shear=False,
                                      ia_bias=(z_arr, b_IA), use_A_ia=False)
    wl_tracer = ccl.WeakLensingTracer(cosmo, dndz=(z_arr, pz))

    cl_GI_1h = ccl.angular_cl(cosmo, wl_tracer, ia_tracer, l_arr,
                              p_of_k_a=pk_GI_1h)
    cl_II_1h = ccl.angular_cl(cosmo, ia_tracer, ia_tracer, l_arr,
                              p_of_k_a=pk_II_1h)

    assert np.all(np.fabs(cl_GI_1h / cl_GI_benchmark - 1) < 0.01)
    assert np.all(np.fabs(cl_II_1h / cl_II_benchmark - 1) < 0.02)

    # Below we are benchmarking the normalization factor of the
    # delta-E power spectrum Eq. (17) in Fortuna et al. 2021.
    # Specifically, the term f_s(z)*<N_s|M>/n_s(z).
    mass_benchmark, norm_benchmark = np.loadtxt(
        './benchmarks/data/IA_halomodel_norm_term.dat', unpack=True)
    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, Omega_k=0,
                          sigma8=0.81, n_s=0.96, h=1.)
    cosmo.compute_sigma()
    Ns_M_mean = sat_gamma_HOD._Ns(mass_benchmark, 1.)
    norm = Ns_M_mean / sat_gamma_HOD.get_normalization(cosmo, 1, hmc=hmc)
    assert np.all(np.fabs(norm / norm_benchmark - 1) < 0.005)



import numpy as np
import os, subprocess, glob, random
import pyccl as ccl


def build_data_dict(stats_arr, prefix):
    """
    Build a data dictionary with columns named according to (k,z) bins, a
    threshold value, and some prefix.

    Assumes that stats_arr has shape: (N_samp, N_thres, N_z, N_kbins)
    """
    # Get no. of points in each dimension.
    N_sam, N_thres, N_z, N_kbins = stats_arr.shape

    # Create dictionary with column names that can be used by ColumnDataSource
    data_dict = {}
    for n in range(N_thres):
        for j in range(N_z):
            for m in range(N_kbins):
                key = "tot_%s_h%d_k%d_z%d" % (prefix, n+1, m+1, j+1)
                data_dict[key] = stats_arr[:,n, j, m]
    return data_dict


def save_hypercube(fname, sample_points):
    """
    Save a Latin hypercube to disk as a text file. Parameter columns will be
    ordered alphabetically by name.

    Parameters
    ----------
    fname : str
        Filename for output file.
    sample_points : dict
        Dictionary containing parameter names (keys) and array of parameter
        values for all sample points (values).
    """
    # Get parameter names and build header
    pnames = sample_points.keys()
    pnames.sort()
    hdr = " ".join(pnames)

    # Build array
    dat = np.column_stack([sample_points[p] for p in pnames])
    np.savetxt(fname, dat, header=hdr, fmt="%4.4e")
    print("Saved hypercube to '%s'." % fname)


def load_hypercube(fname):
    """
    Load a Latin hypercube from disk. Parameter columns will be ordered
    alphabetically by name.

    Parameters
    ----------
    fname : str
        Filename to read sample points from.

    Returns
    -------
    sample_points : dict
        Dictionary containing parameter names (keys) and array of parameter
        values for all sample points (values).
    """
    # Get header
    f = open(fname, 'r')
    hdr = f.readline()[2:-1].split(" ")
    f.close()

    # Load data
    dat = np.loadtxt(fname).T

    # Build dict
    sample_points = {}
    for i in range(len(hdr)):
        sample_points[hdr[i]] = dat[i]
    return sample_points


def generate_latin_hypercube(samples, param_dict, class_root, seed=10):
    """
    Generate a Latin hypercube for a given set of parameters.

    Parameters
    ----------
    samples : int
        Number of samples points to draw from the hypercube.

    param_dict : dict
        Dictionary containing parameter names (keys) and tuple with min/max.
        values (values).

    class_root : str
        Path to directory containing 'class' executable.

    seed : int, optional
        Random seed to use when sampling from hypercube.

    Returns
    -------
    sample_points : dict
        Dictionary containing parameter names (keys) and array of parameter
        values for all sample points (values).
    """
    # Set random seed
    random.seed(seed)

    # Create dictionary to hold sampled parameter values
    sample_points = {}
    for key in param_dict.keys():
        sample_points[key] = np.zeros(samples)
    Ndim = len(param_dict.keys())
    pnames = [key for key in param_dict.keys()]

    # List of indices for each dimension
    l = [range(samples) for j in range(Ndim)]

    # Generate samples until there are no indices left to choose
    for i in range(samples):

        # Randomly choose index and then remove the number that was chosen
        # (Latin hypercubes require at most one item per row and column)
        for j, p in enumerate(pnames):
            pmin, pmax = param_dict[p]
            idx = random.choice(l[j])

            # Get value at this sample point (add 0.5 to idx get bin centroid)
            sample_points[p][i] = pmin + (pmax - pmin) \
                                * (idx + 0.5) / float(samples)
            l[j].remove(idx) # Remove choice from list (sampling w/o replacement)

    return sample_points


def generate_ccl_pspec(sample_points, root, class_data_root, zvals,
                       default_params=None, nonlin=False, mode='std'):
    """
    Generate linear and non-linear power spectra using CCL, for a set of
    points in cosmological parameter space and redshift.

    Parameters
    ----------
    sample_points : dict
        Dictionary containing parameter names (keys) and array of parameter
        values for all sample points (values).

    root : str
        Path of directory in which output files should be stored.

    class_data_root : str
        Root of filenames in which CLASS power spectra were stored.

    zvals : array_like
        Array of redshifts at which the power spectra should be evaluated.

    default_params : dict, optional
        Dictionary of default cosmological parameters, to be used if a
        necessary parameter is not included in sample_points.

    nonlin : bool, optional
        Whether to load non-linear P(k). Default: False.

    mode : str, optional
        Which mode the CLASS run was performed in. Default: 'std'.
    """

    # Get list of parameter names
    pnames = sample_points.keys()
    Nsamp = sample_points[pnames[0]].size

    # Loop over sets of parameters, getting CCL power spectra for each
    for i in range(Nsamp):
        print("Calculating CCL power spectrum %d / %d" % (i, Nsamp))

        # Load the CLASS power spectra to get k bins (lin and NL can differ!)
        class_file = "%s_%05dz1_pk.dat" % (class_data_root, i)
        k_class, pk_class = np.genfromtxt(class_file).T
        k_class *= sample_points['h'][i] # Convert to non-h^-1 units

        if nonlin:
            class_file_nl = "%s_%05dz1_pk_nl.dat" % (class_data_root, i)
            k_class_nl, pk_class_nl = np.genfromtxt(class_file_nl).T
            k_class_nl *= sample_points['h'][i] # Convert to non-h^-1 units

        # Build parameter dictionary
        params = {}
        for p in sample_points.keys():
            # Treat parameters with different naming conventions as special case
            if p == 'Omega_cdm':
                params['Omega_c'] = sample_points[p][i]
            else:
                params[p] = sample_points[p][i]

        # Instantiate CCL Cosmology object with this set of parameters
        params['transfer_function'] = 'boltzmann'
        cosmo = ccl.Cosmology(**params)

        # Loop over redshifts to get linear and nonlinear matter power spectra
        errored = []
        for j, z in enumerate(zvals):
            try:
                a = 1. / (1. + z)
                pk_lin = ccl.linear_matter_power(cosmo, k_class, a)
                if nonlin:
                    pk_nl = ccl.nonlin_matter_power(cosmo, k_class_nl, a)
            except KeyboardInterrupt:
                raise
            except:
                print("--- Error running CLASS")
                print("--- Parameters:", params)
                errored.append(i)
                continue

            # Save datafiles
            if nonlin:
                fname_nl = "%s_nl_%s_%05d_z%d.dat" % (root, mode, i, j+1)
                np.savetxt(fname_nl, np.column_stack((k_class_nl, pk_nl)))
            else:
                fname_lin = "%s_lin_%s_%05d_z%d.dat" % (root, mode, i, j+1)
                np.savetxt(fname_lin, np.column_stack((k_class, pk_lin)))

    # Print runs that errored
    if len(errored) > 0: print("ERRORED:")
    for err in errored:
        print(err)


def generate_class_ini(sample_points, root, nonlinear=False, mnu=False,
                       redshifts=np.arange(0., 3., 0.5)):
    """
    Generate CLASS .ini files for a set of parameters.

    Parameters
    ----------
    sample_points : dict
        Dictionary containing parameter names (keys) and array of parameter
        values for all sample points (values).

    root : str
        Path of directory in which .ini files should be stored.

    nonlinear : bool, optional
        Whether CLASS should return the linear or nonlinear (halofit) power
        spectrum. Default: False.

    mnu : bool, optional
        Whether to include massive neutrinos.

    redshifts : array_like, optional
        Array of redshift values at which P(k) should be calculated. Default:
        Five bins from 0.0 to 2.5 in increments of 0.5.
    """
    # Get user-defined parameter names
    pnames = sample_points.keys()
    Nsamp = sample_points[pnames[0]].size

    # Loop over sample points
    for i in range(Nsamp):
        if i % 10 == 0: print("  Writing CLASS .ini file %d / %d" % (i, Nsamp))

        # Open file for writing
        f = open("%s_%05d.ini" % (root, i), 'w')

        # Write output location into file (will be same as .ini file location)
        f.write('root = %s_%05d\n' % (root, i))

        # Write user-defined cosmo parameters into file
        for p in pnames:
            # Handle commonly-used params that CLASS uses different names for
            if p == 'w0' or p == 'w_0':
                f.write("w0_fld = %e\n" % sample_points[p][i])
            elif p == 'wa' or p == 'w_a':
                f.write("wa_fld = %e\n" % sample_points[p][i])
            else:
                # Generic user-defined parameters
                f.write("%s = %e\n" % (p, sample_points[p][i]))

        # Write output redshifts to file
        f.write("z_pk = %s\n" % (",".join(["%3.3f" % z for z in redshifts])))

        # Write nonlinear switch to file, if specified
        if nonlinear: f.write("non linear = halofit\n")

        # Write curvature parameter
        if 'Omega_k' not in pnames: f.write('Omega_k = 0.0\n')

        # Write neutrino parameters
        N_ur = 2.0328 if mnu else 3.046
        N_ncdm = 1 if mnu else 0
        f.write("N_ur = %f\n" % N_ur)
        f.write("N_ncdm = %f\n" % N_ncdm)
        if mnu: f.write("m_ncdm = 0.06\n")

        # Add various default CLASS settings to file
        # NB: Need to set P_k_max_h/Mpc to a large value if doing halofit
        defaults = "T_cmb = 2.725\n\
#Omega_dcdmdr = 0.0\n\
#Gamma_dcdm = 0.0 \n\
#Omega_fld = 0\n\
#Omega_scf = 0\n\
#a_today = 1.\n\
#YHe = BBN\n\
#recombination = RECFAST\n\
#reio_parametrization = reio_camb\n\
#z_reio = 11.357\n\
#reionization_exponent = 1.5\n\
#reionization_width = 0.5\n\
#helium_fullreio_redshift = 3.5\n\
#helium_fullreio_width = 0.5\n\
#annihilation = 0.\n\
#annihilation_variation = 0.\n\
#annihilation_z = 1000\n\
#annihilation_zmax = 2500\n\
#annihilation_zmin = 30\n\
#annihilation_f_halo = 20\n\
#annihilation_z_halo = 8\n\
#on the spot = yes\n\
#decay = 0.\n\
output = mPk\n\
modes = s\n\
lensing = no\n\
#ic = ad\n\
#P_k_ini type = analytic_Pk\n\
#k_pivot = 0.05\n\
#alpha_s = 0.\n\
P_k_max_h/Mpc = 100.\n\
#l_max_scalars = 8000\n\
#l_max_lss = 1000\n\
headers = yes\n\
format = class\n\
write background = no\n\
write thermodynamics = no\n\
write primordial = no\n\
write parameters = yeap\n\
input_verbose = 1\n\
background_verbose = 1\n\
thermodynamics_verbose = 1\n\
perturbations_verbose = 1\n\
transfer_verbose = 1\n\
primordial_verbose = 1\n\
spectra_verbose = 1\n\
nonlinear_verbose = 1\n\
lensing_verbose = 1\n\
output_verbose = 1\n"
        f.write(defaults)
        f.close()


def run_class(fname_pattern, class_root, precision=False):
    """
    Run CLASS on a set of .ini files.

    Parameters
    ----------
    fname_pattern : str
        Pattern (glob format) for matching the .ini files to be run by CLASS.

    class_root : str
        Directory in which the 'class' executable is stored.

    precision : bool, optional
        Whether to run CLASS in high-precision mode (using the pk_ref.pre
        precision file).

    """
    # Get CLASS .ini files that will be run
    fnames = glob.glob(fname_pattern)
    fnames.sort()

    # Loop over all .ini files
    failed = []
    for i, filename in enumerate(fnames):
        print("CLASS run %d / %d (%s)" % (i+1, len(fnames), filename))

        # Run CLASS and save the output
        cmd = ['%s/class' % class_root, filename]
        if precision: cmd += ['%s/pk_ref.pre' % class_root,]
        try:
            stdout = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            basefile = os.path.splitext('%s' % filename)[0]
            f = open('%s.txt' % basefile, 'w')
            f.write(stdout)
            f.close()
        except KeyboardInterrupt:
            raise
        except:
            raise
            # CLASS run failed
            print("    CLASS run failed. Skipping.")
            failed.append(i)


def load_summary_stats(sample_points, ccl_data_root, class_data_root,
                       thresholds=[5e-5, 1e-4, 5e-4, 1e-3, 5e-3, 1e-2],
                       scale_ranges = [(1e-4, 1e-2), (1e-2, 1e-1), (1e-1, 1e0)],
                       z_vals = ['1', '2', '3', '4', '5', '6'],
                       cache_name=None):
    """
    Calculate summary stats for the deviation between CCL and reference power
    spectra as a function of scale and redshift, for a large number of sample
    points over the cosmological parameter space.

    Parameters
    ----------
    sample_points : dict
        Dictionary containing parameter names (keys) and array of parameter
        values for all sample points (values).

    ccl_data_root : str
        Root of filenames of CCL power spectrum files

    class_data_root : str
        Root of filenames of CLASS power spectrum files. If the string '_nl_'
        is found in class_data_root, this will assume that nonlinear power
        spectra should be loaded.

    """
    # Get dimensions of stats array that will be constructed
    N_samp = sample_points[sample_points.keys()[0]].size
    N_thres = len(thresholds)
    N_z = len(z_vals)
    N_kbins = len(scale_ranges)

    # Check if data were cached
    if cache_name is not None:
        try:
            stats = np.load("%s.npy" % cache_name)
            print("  Loaded '%s' from cache." % cache_name)

            frac_dev = np.load("%s.frac_dev.npy" % cache_name)
            print("  Loaded '%s.frac_dev' from cache." % cache_name)

            k_arr = np.load("%s.k_arr.npy" % cache_name)
            print("  Loaded '%s.k_arr' from cache." % cache_name)

            assert stats.shape == (N_samp, N_thres, N_z, N_kbins)
            return stats, frac_dev, k_arr
        except:
            print("  Cache '%s' not found. Recomputing." % cache_name)
            pass

    # Create array to hold summary statistics, with shape:
    # (N_samp, N_thres, N_z, N_kbins)
    stats = np.zeros((N_samp, N_thres, N_z, N_kbins))
    frac_dev = [[[] for j in range(N_z)] for i in range(N_samp)]
    k_arr = [[[] for j in range(N_z)] for i in range(N_samp)]

    # Loop over sample points in parameter space and calculate summary stats
    for i in range(N_samp):
        print("  Loading power spectra for parameter set %05d" % i)

        # Get Hubble parameter, h, for rescaling CLASS P(k) to Mpc units
        h = sample_points['h'][i]

        # Try to get deviation
        try:
          # Loop over redshift values
          for j in range(N_z):

            # Construct filenames for CCL and CLASS P(k) data files
            fname_ccl = "%s_%05d_z%d.dat" % (ccl_data_root, i, j+1)
            if '_nl_' in class_data_root:
                fname_class = "%s_%05dz%d_pk_nl.dat" % (class_data_root, i, j+1)
            else:
                fname_class = "%s_%05dz%d_pk.dat" % (class_data_root, i, j+1)

            # Load CCL power spectrum data
            pk_ccl_dat = np.loadtxt(fname_ccl)
            ccl_k = pk_ccl_dat[:,0]
            ccl_pk = pk_ccl_dat[:,1]

            # Load CLASS power spectrum data
            pk_class_dat = np.loadtxt(fname_class) #, skiprows=1)
            class_k = pk_class_dat[:,0]
            class_pk = pk_class_dat[:,1] / h**3.

            # Sanity checks
            print(ccl_pk.size, class_pk.size)
            assert ccl_pk.size == class_pk.size

            # Calculate fractional deviation
            frac_dev[i][j] = ccl_pk/class_pk - 1.
            k_arr[i][j] = ccl_k

            # Calculate summary stats in each k bin
            for m in range(N_kbins):
                kmin, kmax = scale_ranges[m]
                idxs = np.logical_and(ccl_k >= kmin, ccl_k < kmax)

                # Calculate deviation statistic, Delta, for a range of
                # threshold values (only values above the threshold are counted)
                for n, thres in enumerate(thresholds):
                    # Calculate deviation statistic
                    dev = np.log10(
                               np.abs(ccl_pk[idxs]/class_pk[idxs] - 1.) / thres)
                    dev[np.where(dev < 0.)] = 0.

                    # Store result in stats array (N_samp, N_thres, N_z, N_kbins)
                    stats[i, n, j, m] = np.sum(dev)
        except:
          raise
          # If there were any failures, set stats to 'nan' for this sample point
          print("  Failed to compute stats for sample %05d." % i)
          stats[i, :, :, :] = np.nan

    # Save to cache file
    if cache_name is not None:
        np.save(cache_name, stats)
        np.save("%s.frac_dev" % cache_name, frac_dev)
        np.save("%s.k_arr" % cache_name, k_arr)
    return stats, np.array(frac_dev), k_arr


def ccl_summary_stats(params,
                      fname_template='../stats/lhs_mpk_err_lin_%05d_z%d.dat',
                      thresholds=[5e-5, 1e-4, 5e-4, 1e-3, 5e-3, 1e-2],
                      scale_ranges = [(1e-4, 1e-2), (1e-2, 1e-1), (1e-1, 1e0)],
                      z_vals = ['1', '2', '3', '4', '5', '6'],
                      cache_name=None):
    """
    Calculate summary stats for the deviation between CCL and reference power
    spectra as a function of scale and redshift, for a large number of sample
    points over the cosmological parameter space.
    """
    # Get dimensions of stats array that will be constructed
    N_samp = params['id'].size
    N_thres = len(thresholds)
    N_z = len(z_vals)
    N_kbins = len(scale_ranges)

    # Check if data were cached
    if cache_name is not None:
        try:
            stats = np.load("%s.npy" % cache_name)
            print("  Loaded '%s' from cache." % cache_name)
            assert stats.shape == (N_samp, N_thres, N_z, N_kbins)
            return stats, params
        except:
            raise

    # Create array to hold summary statistics, with shape:
    # (N_samp, N_thres, N_z, N_kbins)
    stats = np.zeros((N_samp, N_thres, N_z, N_kbins))

    # Loop over sample points in parameter space and calculate summary stats
    for i in range(N_samp):
        trial = params['id'][i]
        print("  Loading CCL power spectra for parameter set %05d" % i)

        # Loop over redshift values
        for j in range(N_z):

            # Load cached CCL power spectrum data
            fname = fname_template % (i, z_vals[j])
            pk_ccl_dat = np.loadtxt(fname, skiprows=1)
            ccl_k = pk_ccl_dat[:,0]
            ccl_pk = pk_ccl_dat[:,1]

            # Calculate summary stats in each k bin
            for m in range(N_kbins):
                kmin, kmax = scale_ranges[m]
                idxs = np.logical_and(ccl_k >= kmin, ccl_k < kmax)

                # Calculate deviation statistic, Delta, for a range of
                # threshold values (only values above the threshold are counted)
                # FIXME: ccl_pk is actually the deviation, which was
                # precomputed somewhere!
                for n, thres in enumerate(thresholds):
                    # Calculate deviation statistic
                    dev = np.log10(np.abs(ccl_pk[idxs]) / thres)
                    dev[np.where(dev < 0.)] = 0.

                    # Store result in stats array (N_samp, N_thres, N_z, N_kbins)
                    stats[i, n, j, m] = np.sum(dev)

    # Save to cache file
    if cache_name is not None:
        np.save(cache_name, stats)

    return stats, params



import pyccl as ccl
import numpy as np
import scipy.integrate
import math
from functools import partial

cosmoin=ccl.Cosmology(
    Omega_c=0.25,
    Omega_b=0.05,
    h=0.7,
    sigma8=0.8,
    n_s=0.96,
    Neff=0,
    m_nu=0.0,
    w0=-1.,
    wa=0.,
    T_CMB=2.7,
    m_nu_type='normal',
    Omega_g=0,
    Omega_k=0,
    transfer_function='bbks',
    matter_power_spectrum='linear')

k_lin=np.logspace(-4., 3., 10000)
a_arr = np.logspace(-1., 0., 10)
knl_arr = np.zeros((len(a_arr),))
for i in range(len(a_arr)):
    a = a_arr[i]
    #I will need linear power at z
    pk_lin_z=ccl.linear_matter_power(cosmoin, k_lin, a)
    interp_pk_lin_z=scipy.interpolate.interp1d(k_lin,pk_lin_z)


    #---kNL prediction----
    [intknl,errintknl]=scipy.integrate.quad(interp_pk_lin_z,min(k_lin),max(k_lin),epsabs=0,epsrel=1e-6)
    print("Check error is small:",errintknl/intknl)
    knlemin2=1./(6*math.pi**2)*intknl
    knl=1./math.sqrt(knlemin2)
    print('knl[1/Mpc]=',knl)
    knl_arr[i] = knl

header_knl="[0] a, [1] k_NL Mpc^-1"

np.savetxt("kNL.txt", np.transpose(np.vstack((a_arr ,knl_arr))), header=header_knl)
